import os
import numpy as np
import pytest
import unittest
from unittest.mock import patch, MagicMock

from resmico import models_fl
from resmico import contig_reader
from resmico import reader

from resmico.contig_reader import ContigInfo

# test/data dir
test_dir = os.path.join(os.path.dirname(__file__))
data_dir = os.path.join(test_dir, 'data')
FEAT_DIR = os.path.join(data_dir, 'preprocess')
INFILE = os.path.join(FEAT_DIR, 'features_binary')

class TestBase(unittest.TestCase):
    def assert_array_equal(self, a, b):
        self.assertIsNone(np.testing.assert_array_equal(a, b))


class TestBinaryDatasetTrain(TestBase):
    def test_select_intervals(self):
        contig_data = [ContigInfo('Contig1', '/tmp/c1', 1000, 0, 0, 0, [], avg_coverage=5),
                       ContigInfo('Contig2', '/tmp/c2', 1000, 0, 0, 0, [(100, 100)], avg_coverage=5),
                       ContigInfo('Contig3', '/tmp/c3', 1000, 0, 0, 0, [(800, 900)], avg_coverage=5)]
        max_len = 500
        for i in range(50):
            intervals = models_fl.BinaryDatasetTrain.select_intervals(contig_data, max_len,
                                                                      translate_short_contigs=True,
                                                                      max_translation_bases=30)
            self.assertTrue(0 <= intervals[0][0] <= 500)
            self.assertTrue(500 <= intervals[0][1] <= 1000)
            self.assertTrue(0 <= intervals[1][0] <= 50)
            self.assertTrue(500 <= intervals[1][1] <= 550)
            self.assertTrue(450 <= intervals[2][0] <= 500, f'Start is {intervals[2][0]}')
            self.assertTrue(500 <= intervals[2][1] <= 1000)

    def test_select_intervals_translate_short(self):
        contig_data = [
            ContigInfo('Contig1', '/tmp/c1', 300, 0, 0, 0, [(200, 210)], avg_coverage=5),
        ]
        max_len = 350
        for i in range(50):
            intervals = models_fl.BinaryDatasetTrain.select_intervals(contig_data, max_len, True, 30)
            if intervals[0][1] - intervals[0][0] < 300:  # contig was truncated to left
                self.assertTrue(0 <= intervals[0][0] <= 150)
                self.assertEqual(300, intervals[0][1])
            else:  # contig will be shifted to left
                self.assertTrue(0 <= intervals[0][0] <= 50)
                self.assertEqual(300 + intervals[0][0], intervals[0][1],
                                 f'Intervals are: {intervals[0][0]}  {intervals[0][1]}')

    @patch('resmico.models_fl.BinaryDatasetTrain.select_intervals')
    def test_contig_selection(self, mock_intervals):
        """
        Make sure that the returned contig features (when using translations) are correct.

        The test mocks reader.read_contigs and BinaryDatasetTrain.select_intervals, and checks if the returned
        (x,y) tuple of BinaryDatasetTrain.__get_item__() is correct.
        """
        for use_cython in [False, True]:
            for cached in [False, True]:
                features = ['num_query_A', 'coverage', 'num_SNPs']
                ctg_reader = contig_reader.ContigReader(FEAT_DIR, features, 1, use_cython)
                c1 = ContigInfo('Contig1', '/tmp/c1', 500, 0, 0, 0, [], avg_coverage=5)
                c2 = ContigInfo('Contig2', '/tmp/c2', 300, 0, 0, 1, [(100, 100)], avg_coverage=5)
                c3 = ContigInfo('Contig3', '/tmp/c3', 1000, 0, 0, 1, [(800, 900)], avg_coverage=5)
                ctg_reader.contigs = [c1, c2, c3]

                contigs_data = []
                st = 0
                for c in ctg_reader.contigs:
                    contig_data = {}
                    for f in features:
                        feature_data = np.arange(start=st, stop=st + c.length, dtype=float)
                        contig_data[f] = feature_data
                        st += c.length
                    contigs_data.append(contig_data)
                ctg_reader.read_contigs = MagicMock(
                    return_value=[contigs_data[0], contigs_data[0], contigs_data[0], contigs_data[1], contigs_data[1],
                                  contigs_data[1], contigs_data[2], contigs_data[2], contigs_data[2]])

                indices = np.arange(len(ctg_reader))
                batch_size = 10
                max_len = 500
                data_gen = models_fl.BinaryDatasetTrain(ctg_reader, indices, batch_size,
                                                        features, max_len,
                                                        num_translations=3,
                                                        max_translation_bases=10, fraq_neg=1.0,
                                                        do_cache=cached, show_progress=False,
                                                        convoluted_size=(lambda x, pad: x),
                                                        pad_to_max_len=False,
                                                        weight_factor=100)
                data_gen.indices.sort()  # indices will now be 0,0,0,1,1,1,2,2,2
                self.assertEqual(9, len(data_gen.indices))  # we have 3 contigs, each translated 3 times
                self.assertEqual([0, 0, 0, 1, 1, 1, 2, 2, 2], data_gen.indices)
                mock_intervals.return_value = [
                    # 1st contig: full contig, cropped to left 100, cropped to left 10
                    (0, 500), (100, 500), (10, 500),
                    (0, 300), (50, 300), (40, 300),  # 2nd contig
                    (500, 1000), (450, 950), (440, 900)  # 3rd contig
                ]

                (x, mask), y, weights = data_gen.__getitem__(0)
                self.assertEqual((batch_size, max_len, len(features)), x.shape)
                # the last zero is just padding
                self.assert_array_equal([0, 0, 0, 1, 1, 1, 1, 1, 1, 0], y)

                # first contig, 1st translation (no translation, full contig)
                for i in range(500):
                    self.assertEqual(i, x[0][i][0])
                    self.assertEqual(500 + i, x[0][i][1])
                    self.assertEqual(1000 + i, x[0][i][2])
                # first contig, 2nd translation (shift 100 to left)
                for i in range(400):
                    self.assertEqual(100 + i, x[1][i][0])
                    self.assertEqual(600 + i, x[1][i][1])
                    self.assertEqual(1100 + i, x[1][i][2])
                for i in range(400, 500):
                    for j in range(3):
                        self.assertEqual(0, x[1][i][j])

                # first contig, 2nd translation (shift 10 to left)
                for i in range(490):
                    self.assertEqual(10 + i, x[2][i][0])
                    self.assertEqual(510 + i, x[2][i][1])
                    self.assertEqual(1010 + i, x[2][i][2])
                for i in range(490, 500):
                    for j in range(3):
                        self.assertEqual(0, x[2][i][j])

                # 2nd contig 1st translation
                for i in range(300):
                    self.assertEqual(1500 + i, x[3][i][0])
                    self.assertEqual(1800 + i, x[3][i][1])
                    self.assertEqual(2100 + i, x[3][i][2])
                for i in range(300, 500):
                    for j in range(3):
                        self.assertEqual(0, x[3][i][j])

                # 2nd contig 2nd translation (truncate 50 bases from the left)
                for i in range(250):
                    self.assertEqual(1550 + i, x[4][i][0])
                    self.assertEqual(1850 + i, x[4][i][1])
                    self.assertEqual(2150 + i, x[4][i][2])
                for i in range(250, 500):
                    for j in range(3):
                        self.assertEqual(0, x[4][i][j])

                # 2nd contig 3rd translation (truncate 40 bases from the left)
                for i in range(260):
                    self.assertEqual(1540 + i, x[5][i][0])
                    self.assertEqual(1840 + i, x[5][i][1])
                    self.assertEqual(2140 + i, x[5][i][2])
                for i in range(260, 500):
                    for j in range(3):
                        self.assertEqual(0, x[5][i][j])

                # 3rd contig 1st translation (last 500 bases of contig 3)
                for i in range(500):
                    self.assertEqual(2900 + i, x[6][i][0])
                    self.assertEqual(3900 + i, x[6][i][1])
                    self.assertEqual(4900 + i, x[6][i][2])

                # 3rd contig 2nd translation (bases 450 to 950 of contig 3)
                for i in range(500):
                    self.assertEqual(2850 + i, x[7][i][0])
                    self.assertEqual(3850 + i, x[7][i][1])
                    self.assertEqual(4850 + i, x[7][i][2])

                # 3rd contig 3rd translation (bases 440 to 950 of contig 3)
                for i in range(460):
                    self.assertEqual(2840 + i, x[8][i][0])
                    self.assertEqual(3840 + i, x[8][i][1])
                    self.assertEqual(4840 + i, x[8][i][2])

                # padding
                for i in range(500):
                    for j in range(3):
                        self.assertEqual(0, x[9][i][j])

    def test_gen_train_data(self):
        for cached in [False, True]:
            ctg_reader = contig_reader.ContigReader(FEAT_DIR, reader.feature_names, 1, False)
            indices = np.arange(len(ctg_reader))
            batch_size = 10
            num_translations = 1
            data_gen = models_fl.BinaryDatasetTrain(ctg_reader, indices, batch_size,
                                                    reader.feature_names, 500,
                                                    num_translations=1,
                                                    max_translation_bases=0, fraq_neg=1.0,
                                                    do_cache=cached, show_progress=False,
                                                    convoluted_size=(lambda x, pad: x),
                                                    pad_to_max_len=False,
                                                    weight_factor=1000)
            data_gen.translate_short_contigs = False  # so that we know which interval is selected
            # set these to -1 in order to enforce NOT swapping A/T and G/C (for data enhancement)
            data_gen.pos_A = data_gen.pos_ref = data_gen.pos_C = -1
            # unshuffle the indices, so that we can make assertions about the returned data
            data_gen.indices = [0, 1]
            self.assertEqual(1, len(data_gen))
            (train_data, mask), y, weights = data_gen[0]
            # even if we only have 2 samples, the remaining are filled with zero to reach the desired batch size
            self.assertEqual(batch_size, len(train_data))

            self.assert_array_equal(weights, np.array([0.25, 0.25, 1, 1, 1, 1, 1, 1, 1, 1]))

            expected_y = np.zeros(batch_size)
            expected_y[0] = 1
            self.assert_array_equal(y, expected_y)

            # # train_data[0][0] - first position in first contig, train_data[0][5] 5th position in 1st contig
            self.assert_array_equal(train_data[0][0][0:6], np.array([1, 0, 0, 0, 2, 1]))
            self.assert_array_equal(train_data[0][5][0:6], np.array([1, 0, 0, 0, 0, 0]))

            # # train_data[1][0] - first position in 2nd contig, train_data[1][5] 5th position in 2nd contig
            self.assert_array_equal(train_data[1][0][0:6], np.array([1, 0, 0, 0, 1, 1]))
            self.assert_array_equal(train_data[1][5][0:6], np.array([1, 0, 0, 0, 0, 0]))


class TestBinaryDatasetEval(TestBase):
    bytes_per_base = 10 + sum(  # 10 is the overhead also added in Models_Fl.BinaryDataEval
        [np.dtype(ft).itemsize for ft in reader.feature_np_types])

    def test_batching_one_per_batch(self):
        ctg_reader = contig_reader.ContigReader(FEAT_DIR, reader.feature_names, 1, False)
        ctg_reader.contigs = [ContigInfo('Contig1', '/tmp/c1', 1000, 0, 0, 0, [], avg_coverage=5),
                              ContigInfo('Contig2', '/tmp/c2', 1000, 0, 0, 0, [], avg_coverage=5),
                              ContigInfo('Contig3', '/tmp/c3', 1000, 0, 0, 0, [], avg_coverage=5)]
        indices = np.arange(len(ctg_reader))

        gpu_memory_bytes = 1010 * self.bytes_per_base
        eval_data = models_fl.BinaryDatasetEval(ctg_reader, indices, reader.feature_names,
                                                500, 250, gpu_memory_bytes,
                                                False, False, convoluted_size=(lambda x, pad: x),
                                                pad_to_max_len=False)
        self.assertEqual(3, len(eval_data.chunk_counts))
        for i in range(len(eval_data.chunk_counts)):
            self.assertEqual(1, len(eval_data.chunk_counts[i]))
            self.assertEqual(3, eval_data.chunk_counts[i][0])

    def test_batching_multiple_per_batch(self):
        ctg_reader = contig_reader.ContigReader(FEAT_DIR, reader.feature_names, 1, False)
        ctg_reader.contigs = [
            ContigInfo('Contig1', INFILE, 500, 0, 246, 0, [], avg_coverage=5),
            ContigInfo('Contig2', INFILE, 500, 246, 183, 0, [], avg_coverage=5),
            ContigInfo('Contig3', INFILE, 500, 0, 246, 0, [], avg_coverage=5)]
        indices = np.arange(len(ctg_reader))
        gpu_memory_bytes = 1600 * self.bytes_per_base
        eval_data = models_fl.BinaryDatasetEval(ctg_reader, indices, reader.feature_names, 250,
                                                200, gpu_memory_bytes, False, False,
                                                convoluted_size=(lambda x, pad: x),
                                                pad_to_max_len=False)
        # check that Contig1 and Contig2 are in the first batch (with 3 chunks each) and Contig3 is in the second batch
        # (also with 3 chunks)
        # 1st batch, 2 contigs, 3 chunks each
        self.assertEqual(2, len(eval_data.chunk_counts))
        self.assertEqual(2, len(eval_data.chunk_counts[0]))
        self.assertEqual(3, eval_data.chunk_counts[0][0])
        self.assertEqual(3, eval_data.chunk_counts[0][1])
        # 2nd batch, 1 contig, 3 chunks
        self.assertEqual(1, len(eval_data.chunk_counts[1]))
        self.assertEqual(3, eval_data.chunk_counts[1][0])

        (x, _), _ = eval_data[0]
        self.assertEqual(6, len(x))
        (x, _), _ = eval_data[1]
        self.assertEqual(3, len(x))

    def test_gen_eval_data(self):
        for cached in [False, True]:
            ctg_reader = contig_reader.ContigReader(FEAT_DIR, reader.feature_names, 1, False)
            indices = np.arange(len(ctg_reader))
            eval_data = models_fl.BinaryDatasetEval(ctg_reader, indices, reader.feature_names, 500,
                                                    250, 1e6, cached, False,
                                                    convoluted_size=(lambda x, pad: x),
                                                    pad_to_max_len=False)
            self.assertEqual(1, len(eval_data))
            self.assertEqual(2, len(eval_data.batch_list[0]))
            (x, _), _ = eval_data[0]
            self.assert_array_equal(x[0][0][0:6], np.array([1, 0, 0, 0, 2, 1]))
            self.assert_array_equal(x[0][5][0:6], np.array([1, 0, 0, 0, 0, 0]))

            self.assert_array_equal(x[1][0][0:6], np.array([1, 0, 0, 0, 1, 1]))
            self.assert_array_equal(x[1][5][0:6], np.array([1, 0, 0, 0, 0, 0]))

    def test_gen_eval_data_short_window(self):
        ctg_reader = contig_reader.ContigReader(FEAT_DIR, reader.feature_names, 1, False)
        indices = np.arange(len(ctg_reader))
        window = 50
        eval_data = models_fl.BinaryDatasetEval(ctg_reader, indices, reader.feature_names,
                                                window, 30, 1e6, False, False,
                                                convoluted_size=(lambda x, pad: x),
                                                pad_to_max_len=False)
        feature_count = len(eval_data.expanded_feature_names)

        self.assertEqual(1, len(eval_data))  # one batch total
        self.assertEqual(2, len(eval_data.batch_list[0]))  # the one batch has 2 contigs
        self.assertEqual([16, 16], eval_data.chunk_counts[0])  # each of the contigs of length 500 has 16 chunks

        total_chunks = 16 + 16  # 16 for the first contig of length 500, 16 for the 2nd contig of length 500
        (x, _), _ = eval_data[0]
        self.assertEqual((total_chunks, window, feature_count), x.shape)

        # check the first 6 features in the 0th and 5th positions of the first chunk in first contig (reference_A/C/G/T,
        # coverage, num_query_A)
        (x, _), _ = eval_data[0]
        self.assertIsNone(
            np.testing.assert_array_equal(x[0][0][0:6], np.array([1, 0, 0, 0, 2, 1])))
        self.assertIsNone(
            np.testing.assert_array_equal(x[0][5][0:6], np.array([1, 0, 0, 0, 0, 0])))

        # check the first 6 features in the 0th and 5th positions of first chunk in 2nd contig
        self.assertIsNone(
            np.testing.assert_array_equal(x[16][0][0:6], np.array([1, 0, 0, 0, 1, 1])))
        self.assertIsNone(
            np.testing.assert_array_equal(x[16][5][0:6], np.array([1, 0, 0, 0, 0, 0])))

    def test_group(self):
        ctg_reader = contig_reader.ContigReader(FEAT_DIR, reader.feature_names, 1, False)
        indices = np.arange(len(ctg_reader))
        total_memory_bytes = 1e6
        eval_data = models_fl.BinaryDatasetEval(ctg_reader, indices, reader.feature_names, 50, 30,
                                                total_memory_bytes, False, False,
                                                convoluted_size=(lambda x, pad: x),
                                                pad_to_max_len=False)
        (x, _), _ = eval_data[0]
        self.assertEqual(32, len(x))

        y = np.zeros(32)
        self.assertIsNone(
            np.testing.assert_array_equal(np.array([0, 0]), eval_data.group(y)))

        y = np.zeros(32)
        y[5] = 1
        self.assertIsNone(
            np.testing.assert_array_equal(np.array([1, 0]), eval_data.group(y)))
        y[0:15] = 1
        self.assertIsNone(
            np.testing.assert_array_equal(np.array([1, 0]), eval_data.group(y)))

        y[16] = 1
        self.assertIsNone(
            np.testing.assert_array_equal(np.array([1, 1]), eval_data.group(y)))

    def test_group_two_batches(self):
        ctg_reader = contig_reader.ContigReader(FEAT_DIR, reader.feature_names, 1, False)
        indices = np.arange(len(ctg_reader))
        eval_data = models_fl.BinaryDatasetEval(ctg_reader, indices, reader.feature_names,
                                                50, 30, 500, False, False,
                                                convoluted_size=(lambda x, pad: x),
                                                pad_to_max_len=False)

        self.assertEqual(2, len(eval_data))
        (x, _), _ = eval_data[0]
        self.assertEqual(16, len(x))
        (x, _), _ = eval_data[1]
        self.assertEqual(16, len(x))

        y = np.zeros(32)
        self.assertIsNone(
            np.testing.assert_array_equal(np.array([0, 0]), eval_data.group(y)))

        y = np.zeros(32)
        y[5] = 1
        self.assertIsNone(
            np.testing.assert_array_equal(np.array([1, 0]), eval_data.group(y)))
        y[0:15] = 1
        self.assertIsNone(
            np.testing.assert_array_equal(np.array([1, 0]), eval_data.group(y)))

        y[16] = 1
        self.assertIsNone(
            np.testing.assert_array_equal(np.array([1, 1]), eval_data.group(y)))

    def test_gen_eval_data_cached(self):
        ctg_reader = contig_reader.ContigReader(FEAT_DIR, reader.feature_names, 1, False)
        indices = np.arange(len(ctg_reader))
        eval_data = models_fl.BinaryDatasetEval(ctg_reader, indices, reader.feature_names, 500,
                                                250, 1e6, True, False,
                                                convoluted_size=(lambda x, pad: x),
                                                pad_to_max_len=False)
        self.assertEqual(1, len(eval_data))
        self.assertEqual(2, len(eval_data.batch_list[0]))
        (x, _), _ = eval_data[0]

        self.assertIsNone(
            np.testing.assert_array_equal(x[0][0][0:6], np.array([1, 0, 0, 0, 2, 1])))
        self.assertIsNone(
            np.testing.assert_array_equal(x[0][5][0:6], np.array([1, 0, 0, 0, 0, 0])))

        self.assertIsNone(
            np.testing.assert_array_equal(x[1][0][0:6], np.array([1, 0, 0, 0, 1, 1])))
        self.assertIsNone(
            np.testing.assert_array_equal(x[1][5][0:6], np.array([1, 0, 0, 0, 0, 0])))


class TestResmico(unittest.TestCase):
    def setUp(self):
        args = MagicMock()
        args.n_hid = 50
        args.net_type = 'cnn_resnet'
        args.filters = 16
        args.features = ['ref_base']
        args.num_blocks = 4
        args.ker_size = 5
        args.lr_init = 1e-3
        self.args = args

    def test_convolved_output_size(self):
        model = models_fl.Resmico(self.args)
        self.assertIsNotNone(model.convoluted_size)
        self.assertEqual(19, model.convoluted_size(512, False))
        self.assertEqual(83, model.convoluted_size(1024, False))
        # (((1071-9-8-5)//2-4-31-4)//2-4-31-4)//2-4-8
        self.assertEqual(89, model.convoluted_size(1071, True))



import argparse
import json
import os
import logging
from shutil import copyfile
import sys

import numpy as np

from resmico import contig_reader
from resmico import models_fl as Models
from resmico.commands import arguments


def get_toc_name(contig_info: Models.ContigInfo, is_chunked: bool, base_dir: str):
    return os.path.join(os.path.dirname(os.path.join(base_dir, contig_info.file)), 'toc' + (
        '_chunked' if is_chunked else ''))


def get_stats_name(fname: Models.ContigInfo, base_dir: str):
    return os.path.join(os.path.dirname(os.path.join(base_dir, fname)), 'stats')


def write_stats(prev_file, feature_files_path, sum_entropy, sum2_entropy, sum_gc_percent, sum2_gc_percent, sum_coverage,
                sum2_coverage, all_count):
    stats_fname = get_stats_name(prev_file, feature_files_path)
    stats_file = open(stats_fname)
    stats = json.load(stats_file)
    stats_file.close()
    copyfile(stats_fname, stats_fname + '_old')
    stats['seq_window_entropy'] = {'sum': sum_entropy, 'sum2': sum2_entropy}
    stats['seq_window_perc_gc'] = {'sum': sum_gc_percent, 'sum2': sum2_gc_percent}
    stats['coverage'] = {'sum': int(sum_coverage), 'sum2': int(sum2_coverage)}
    stats['all_count'] = int(all_count)
    stats_file = open(stats_fname, 'w')
    json.dump(stats, stats_file, indent=2)
    stats_file.close()


def main():
    parser = argparse.ArgumentParser(description='Add stats')
    arguments.add_common_args(parser)

    args = parser.parse_args(sys.argv[1:])

    logging.basicConfig(format='%(asctime)s - %(message)s', level=logging._nameToLevel[args.log_level.upper()])

    args.features = ['coverage', 'seq_window_perc_gc', 'seq_window_entropy']

    logging.info('Loading contig metadata data...')
    reader = contig_reader.ContigReader(args.feature_files_path, args.features, args.n_procs,
                                        args.no_cython)

    logging.info(f'Loaded metadata for {len(reader.contigs)} contigs. ')
    logging.info('Loading actual contig data...')
    if len(reader.contigs) == 0:
        logging.info('Nothing to do')
        exit(0)

    contig_data = reader.read_contigs(reader.contigs, return_raw=True)
    logging.info('All contigs loaded in memory. Updating toc/stats files')

    prev_file = ''
    toc_new = None
    toc_header = 'Contig\tLengthBases\tMisassemblyCnt\tSizeBytes\tBreakingPoints\tAvgCoverage\n'
    sum_entropy: float = 0
    sum_gc_percent: float = 0
    sum2_entropy: float = 0
    sum2_gc_percent: float = 0
    all_count: float = 0
    sum_coverage: int = 0
    sum2_coverage: int = 0

    for contig_info, contig_data in zip(reader.contigs, contig_data):
        if prev_file != contig_info.file:  # new file starting
            if toc_new is not None:
                toc_new.close()
            toc_fname = get_toc_name(contig_info, args.chunks, args.feature_files_path)
            copyfile(toc_fname, toc_fname + '_old')
            toc_new = open(toc_fname, 'w')
            toc_new.write(toc_header)
            # write previous data, if any
            if prev_file != '' and not args.chunks:  # stats are computed on the full data (not the contig chunks)
                write_stats(prev_file, args.feature_files_path, sum_entropy, sum2_entropy, sum_gc_percent,
                            sum2_gc_percent, sum_coverage, sum2_coverage, all_count)
                sum_entropy = 0
                sum_gc_percent = 0
                sum2_entropy = 0
                sum2_gc_percent = 0
                all_count = 0
                sum_coverage = 0
                sum2_coverage = 0
            prev_file = contig_info.file

        breakpoints = '-' if len(contig_info.breakpoints) == 0 else ",".join(
            str(b[0]) + "-" + str(b[1]) for b in contig_info.breakpoints)
        toc_new.write(f'{contig_info.name}\t{contig_info.length}\t{contig_info.misassembly}\t'
                      f'{contig_info.size_bytes}\t{breakpoints}\t'
                      f'{round(np.average(contig_data["coverage"]), 5)}\n')

        all_count += len(contig_data['coverage'])
        sum_entropy += np.sum(contig_data['seq_window_entropy'])
        sum2_entropy += np.sum(contig_data['seq_window_entropy'] ** 2)
        sum_gc_percent += np.sum(contig_data['seq_window_perc_gc'])
        sum2_gc_percent += np.sum(contig_data['seq_window_perc_gc'] ** 2)
        sum_coverage += np.sum(contig_data['coverage'])
        sum2_coverage += np.sum(contig_data['coverage'] ** 2)

    # write leftover data
    write_stats(prev_file, args.feature_files_path, sum_entropy, sum2_entropy, sum_gc_percent,
                sum2_gc_percent, sum_coverage, sum2_coverage, all_count)

    logging.info('Done.')

if __name__ == '__main__':
    main()



from __future__ import print_function
from pkg_resources import resource_filename
import os
import argparse
import logging

from resmico import evaluate
from resmico.commands import arguments


# functions
def get_desc():
    desc = 'Evaluate model'
    return desc


def parse_args(test_args=None, subparsers=None):
    desc = get_desc()
    epi = """DESCRIPTION:
    Evaluate trained resmico model or the one generated by `resmico train`.

    Contig features used for evaluation must have been generated with resmico-sm.
    """
    if subparsers:
        parser = subparsers.add_parser('evaluate', description=desc, epilog=epi,
                                       formatter_class=argparse.RawTextHelpFormatter)
    else:
        parser = argparse.ArgumentParser(description=desc, epilog=epi,
                                         formatter_class=argparse.RawTextHelpFormatter)

    # default trained model
    pkg_model = resource_filename('resmico', 'model/resmico.h5')
    parser_g1 = parser.add_argument_group('Evaluation-specific arguments')
    parser_g1.add_argument('--model', default=pkg_model, type=str,
                        help='Location of the saved deep learning model (default: %(default)s)')
    parser_g1.add_argument('--batch-size', default=300, type=int,
                        help='Batch size (default: %(default)s)')
    parser_g1.add_argument('--embeddings', action='store_true', default=False,
                        help='Produce embeddings for an intermidiate layer (default: %(default)s)')
    parser_g1.add_argument('--emb-num', default=10000, type=int,
                        help='Number of randomly selected contigs (default: %(default)s)')
    parser_g1.add_argument('--emb-ind', default=0, type=int,
                        help='Layer index to produce embedding (default: %(default)s)')
    parser_g1.add_argument('--verify-insert-size', action='store_true', default=False,
                        help='Check if the insert size distribution is similar to the n9k-train dataset (default: %(default)s)')
    arguments.add_common_args(parser)

    # running test args
    if test_args:
        args = parser.parse_args(test_args)
        return args

    return parser


def main(args=None):
    if args is None:
        args = parse_args()

    logging.basicConfig(format='%(asctime)s - %(message)s', level=logging._nameToLevel[args.log_level.upper()])

    print()
    print(args)
    print()
    evaluate.main(args)


if __name__ == '__main__':
    pass



import os
import pytest
import logging

# test/data dir
test_dir = os.path.join(os.path.dirname(__file__))
data_dir = os.path.join(test_dir, 'data')


# tests
def test_helps(script_runner):
    ret = script_runner.run('resmico', 'evaluate', '-h')
    assert ret.success

def test_train(tmpdir, script_runner):
    save_path = tmpdir.mkdir('save_dir')
    input_path = os.path.join(data_dir, 'n10', 'features')
    model_path = os.path.join(data_dir, 'model_n10_e2.h5')
    ret = script_runner.run('resmico', 'evaluate',
                            '--model', model_path,
                            '--save-path', str(save_path),
                            '--feature-files-path', input_path)
    assert ret.success, ret.print()
    



import logging
import os
from pathlib import Path
import time
import atexit

import numpy as np
import pandas as pd
import tensorflow as tf
from tensorflow.keras.models import load_model
from tensorflow.keras.models import Model
from sklearn.metrics import recall_score, average_precision_score

from resmico import contig_reader
from resmico import models_fl as Models
from resmico import utils


def predict_bin_data(model: tf.keras.Model, num_gpus: int, args):
    start = time.time()

    is_fixed_length = model.layers[0].input_shape[0][1] is not None  
    convoluted_size = Models.construct_convolution_lambda(model)
#     else:  # when not padding, the convoluted size is unused
#         convoluted_size = lambda len, pad: 0

    logging.info('Loading contig data...')
    reader = contig_reader.ContigReader(args.feature_files_path, args.features, args.n_procs,
                                        args.no_cython, args.stats_file, args.min_contig_len,
                                        min_avg_coverage=args.min_avg_coverage)

    if args.val_ind_f:
        eval_idx = list(pd.read_csv(args.val_ind_f)['val_ind'])
        logging.info(f'Using {len(eval_idx)} indices in {args.val_ind_f} for prediction')
    elif args.embeddings:
        all_idx = np.arange(len(reader))
        logging.info(f'Using 10k samples for embeddings')
        np.random.shuffle(all_idx)
        eval_idx = all_idx[:args.emb_num]
    else:
        logging.info(f'Using all indices for prediction')
        eval_idx = np.arange(len(reader))

    predict_data = Models.BinaryDatasetEval(reader, eval_idx, args.features, args.max_len, max(250, args.max_len-500), 
                                            int(args.gpu_eval_mem_gb * 1e9 * 0.8), cache_results=False,
                                            show_progress=True, convoluted_size=convoluted_size,
                                            pad_to_max_len=is_fixed_length, batch_size=args.batch_size)

    eval_data_y = np.array([0 if reader.contigs[idx].misassembly == 0 else 1 for idx in predict_data.indices])

    # convert the slow Keras predict_data of type Sequence to a tf.data object
    data_iter = lambda: (s for s in predict_data)

    predict_data_tf = tf.data.Dataset.from_generator(
        data_iter,
        output_signature=(
            # first dimension is batch size, second is contig length, third is number of features
            (tf.TensorSpec(shape=(None, None, len(predict_data.expanded_feature_names)), dtype=tf.float32),
             # first dimension is batch size, second is contig length (no third dimension,
             # as all features are masked the same way)
             tf.TensorSpec(shape=(None, None), dtype=tf.bool)),
            tf.TensorSpec(shape=(None), dtype=tf.bool)
        ))

    predict_data_tf = predict_data_tf.prefetch(4 * num_gpus)
    options = tf.data.Options()
    options.experimental_distribute.auto_shard_policy = tf.data.experimental.AutoShardPolicy.DATA
    predict_data_tf = predict_data_tf.with_options(options)  # avoids Tensorflow ugly console barf

    eval_data_flat_y = model.predict(x=predict_data_tf,
                                     workers=args.n_procs,
                                     use_multiprocessing=True,
                                     max_queue_size=max(args.n_procs, 10),
                                     verbose=0)
    eval_data_predicted_min = predict_data.group(eval_data_flat_y, min)
    eval_data_predicted_mean = predict_data.group(eval_data_flat_y, np.mean)
    eval_data_predicted_std = predict_data.group(eval_data_flat_y, np.std)
    eval_data_predicted_max = predict_data.group(eval_data_flat_y, max)
    eval_data_predicted_score = predict_data.group(eval_data_flat_y, max) #'sma' 'prob'

    auc_val = average_precision_score(eval_data_y, eval_data_predicted_score)
    recall1_val = recall_score(eval_data_y, eval_data_predicted_score > 0.5, pos_label=1)
    recall0_val = recall_score(eval_data_y, eval_data_predicted_score > 0.5, pos_label=0)
    logging.info(f'Prediction scores: aucPR: {auc_val} - recall1: {recall1_val} - recall0: {recall0_val}')
    
    duration = time.time() - start
    logging.info(f'Prediction done in {duration:.0f}s.')

    # saving output
    if not os.path.isdir(args.save_path):
        os.makedirs(args.save_path)
    out_file = os.path.join(args.save_path, args.save_name + '.csv')    
    if args.embeddings:
        middle_output = Model(inputs=model.input, outputs=model.layers[args.emb_ind].output)
        eval_data_emb = middle_output.predict(x=predict_data_tf,
                                                 workers=args.n_procs,
                                                 use_multiprocessing=True,
                                                 max_queue_size=max(args.n_procs, 10),
                                                 verbose=1)

        eval_data_emb = predict_data.group_emb(eval_data_emb, np.mean)
        
        with open(out_file, 'w') as outF:
            outF.write('cont_name,length,label,embedding,score,min,mean,std,max\n')
            for idx in range(len(eval_idx)):
                contig = reader.contigs[predict_data.indices[idx]]
                outF.write(f'{os.path.join(os.path.dirname(contig.file), contig.name)},{contig.length},'
                           f'{contig.misassembly},{eval_data_emb[idx]},{eval_data_predicted_score[idx]},'
                           f'{eval_data_predicted_min[idx]},{eval_data_predicted_mean[idx]},'
                           f'{eval_data_predicted_std[idx]},{eval_data_predicted_max[idx]}\n')        
    else:
        with open(out_file, 'w') as outF:
            outF.write('cont_name,length,label,score,min,mean,std,max\n')
            for idx in range(len(eval_idx)):
                contig = reader.contigs[predict_data.indices[idx]]
                outF.write(f'{os.path.join(os.path.dirname(contig.file), contig.name)},{contig.length},'
                           f'{contig.misassembly},{eval_data_predicted_score[idx]},'
                           f'{eval_data_predicted_min[idx]},{eval_data_predicted_mean[idx]},'
                           f'{eval_data_predicted_std[idx]},{eval_data_predicted_max[idx]}\n')

    logging.info(f'Predictions saved to: {out_file}')

def verify_insert_size(args):
    
    logging.info('Loading contig data...')
    reader = contig_reader.ContigReader(args.feature_files_path, ['mean_insert_size_Match'], args.n_procs, args.chunks,
                                        args.no_cython, args.stats_file, args.min_len,
                                        min_avg_coverage=args.min_avg_coverage)

    contig_data = [reader.contigs[i] for i in range(0, len(reader), 10)] ##10% of the data
    
    features_data = reader.read_contigs(contig_data, return_raw=True)

    insert_size_data = []
    for cont in features_data:
        insert_size_data.extend(cont['mean_insert_size_Match'])

        
    low_train, high_train = 178, 372 #0.05 and 0.95 quantiles of the n9k-train dataset
     
    lowq = np.nanquantile(insert_size_data, 0.06)
    hiq = np.nanquantile(insert_size_data, 0.94)
    if lowq >= low_train and  hiq <= high_train:
        if np.nanquantile(insert_size_data, 0.05) >= low_train and np.nanquantile(insert_size_data, 0.95) <= high_train:
            logging.info('The insert size distribution lies inside the training one. It is safe to apply ResMiCo.')
        else:
            logging.info('The insert size distribution lies close to the border of the training one. ResMiCo can be applied.')
    else:
        logging.info('The insert size distribution is dissimilar to the training data. ResMiCo predictions are not reliable.')
        
def main(args):
    """
    Main interface
    """
    # insert size assessment
    if args.verify_insert_size:
        verify_insert_size(args)
        exit()
    # TODO: remove GlobalMaskedMaxPooling1D, once annotation kicks in
    custom_obj = {'class_recall_0': utils.class_recall_0, 'class_recall_1': utils.class_recall_1,
                  'GlobalMaskedMaxPooling1D': Models.GlobalMaskedMaxPooling1D}
    # model loading
    if not os.path.exists(args.model):
        raise IOError(f'Cannot find {args.model}')
    strategy = tf.distribute.MirroredStrategy()
    logging.info(f'Number of devices: {strategy.num_replicas_in_sync}')
    logging.info(f'Loading model: {args.model}')
    with strategy.scope():
        model = load_model(args.model, custom_objects=custom_obj)
    logging.info('Model loaded')
    # predict
    predict_bin_data(model, strategy.num_replicas_in_sync, args)
#     utils.test_calibprob_vc_max(model, strategy.num_replicas_in_sync, args)
    # exit
#     atexit.register(strategy._extended._collective_ops._pool.close)


if __name__ == '__main__':
    pass



import os
import numpy as np
import unittest

from resmico import contig_reader
from resmico import reader

test_dir = os.path.join(os.path.dirname(__file__))
data_dir = os.path.join(test_dir, 'data')
FEAT_DIR = os.path.join(data_dir, 'preprocess')
INFILE = os.path.join(FEAT_DIR, 'features_binary')

class TestReadContig(unittest.TestCase):
    def test_read_from_file(self):
        input_file = open(INFILE, 'rb')

        result = contig_reader._read_contig_data(input_file, reader.feature_names)
        self.assertEqual(500, len(result['ref_base_A']))
        self.assertIsNone(
            np.testing.assert_array_equal(np.array([1] * 498 + [0, 0]), result['ref_base_A']))
        self.assertIsNone(
            np.testing.assert_array_equal(np.array([0] * 498 + [1, 1]), result['ref_base_C']))
        self.assertTrue(not np.any(result['ref_base_G']))
        self.assertTrue(not np.any(result['ref_base_T']))

        coverage = result['coverage']
        for pos in range(0, len(result['ref_base_A'])):
            self.assertEqual(2 if 420 <= pos < 425 or pos < 5 else 0, coverage[pos])
            self.assertEqual(0, result['num_discordant'][pos])
            self.assertEqual(1 if pos == 0 else 0.5 if pos < 5 else 0, result['num_query_A'][pos])
            self.assertEqual(0.5 if 420 <= pos < 425 else 0, result['num_query_C'][pos])
            self.assertEqual(0.5 if 0 < pos < 5 else 0, result['num_query_G'][pos])
            self.assertEqual(0.5 if 420 <= pos < 425 else 0, result['num_query_T'][pos])

            self.assertIsNone(np.testing.assert_equal(-28 if pos == 0 else 0 if pos < 5 else np.nan,
                                                      result['min_al_score_Match'][pos]))
            self.assertIsNone(np.testing.assert_equal(0 if pos < 5 else np.nan, result['max_al_score_Match'][pos]))

            self.assertIsNone(
                np.testing.assert_array_equal(425 if pos < 5 else np.nan, result['min_insert_size_Match'][pos]))
            self.assertTrue(result['mean_insert_size_Match'][pos] == 425 and pos < 5 or pos >= 5 and np.isnan(
                result['mean_insert_size_Match'][pos]))
            self.assertTrue(result['stdev_insert_size_Match'][pos] == 0 and pos == 0 or pos > 0 and np.isnan(
                result['stdev_insert_size_Match'][pos]))
            self.assertIsNone(
                np.testing.assert_array_equal(425 if pos < 5 else np.nan, result['max_insert_size_Match'][pos]))

            self.assertIsNone(np.testing.assert_array_equal(6 if pos < 5 else np.nan, result['min_mapq_Match'][pos]))
            self.assertIsNone(np.testing.assert_array_equal(7 if pos == 0 else 6 if pos < 5 else np.nan,
                                                            result['max_mapq_Match'][pos]))
            self.assertIsNone(np.testing.assert_array_equal(np.nan if pos >= 5 else 6.5 if pos == 0 else 6,
                                                            result['mean_mapq_Match'][pos]))
            self.assertEqual(0.5 if 0 < pos < 5 else 1 if 420 <= pos < 425 else 0, result['num_proper_SNP'][pos])
            self.assertEqual(0 if pos < 498 else 25 if pos == 498 else 50, result['seq_window_perc_gc'][pos])
            self.assertAlmostEqual(0 if pos < 498 else 0.811278 if pos == 498 else 1, result['seq_window_entropy'][pos],
                                   delta=1e-4)

    def test_normalize_zero_mean_one_stdev(self):
        input_file = open(INFILE, 'rb')
        old_result = contig_reader._read_contig_data(input_file, reader.feature_names)

        c_reader = contig_reader.ContigReader(os.path.split(INFILE)[0],
                                              reader.feature_names, 1)
        for fname in reader.float_feature_names:
            c_reader.means[fname] = 0
            c_reader.stdevs[fname] = 1

        result = c_reader._read_and_normalize(c_reader.contigs[0])

        for fname in reader.float_feature_names:
            # replace NANs with 0, as that's what the normalization in ContigReader does
            nan_pos = np.isnan(old_result[fname])
            old_result[fname][nan_pos] = 0
            self.assertIsNone(np.testing.assert_array_equal(old_result[fname], result[fname]))

    def test_normalize_zero_mean_two_stdev(self):
        input_file = open(INFILE, 'rb')
        old_result = contig_reader._read_contig_data(input_file, reader.feature_names)

        ctg_reader = contig_reader.ContigReader(os.path.split(INFILE)[0],
                                                reader.feature_names, 1)
        for fname in reader.float_feature_names:
            ctg_reader.means[fname] = 0
            ctg_reader.stdevs[fname] = 2

        result = ctg_reader._read_and_normalize(ctg_reader.contigs[0])

        for fname in reader.float_feature_names:
            # replace NANs with 0, as that's what the normalization in ContigReader does
            nan_pos = np.isnan(old_result[fname])
            old_result[fname][nan_pos] = 0
            self.assertIsNone(np.testing.assert_array_equal(old_result[fname] / 2, result[fname]))

    def test_read_toc(self):
        ctg_reader = contig_reader.ContigReader(FEAT_DIR,
                                                [reader.feature_names[0],
                                                 reader.feature_names[1],
                                                 reader.feature_names[3]],
                                                process_count=1)

        self.assertEqual(2, len(ctg_reader.contigs))

        self.assertEqual(500, ctg_reader.contigs[0].length)
        self.assertEqual(500, ctg_reader.contigs[1].length)
        self.assertEqual(1, ctg_reader.contigs[0].misassembly)
        self.assertEqual(0, ctg_reader.contigs[1].misassembly)
        self.assertEqual(256, ctg_reader.contigs[0].size_bytes)
        self.assertEqual(196, ctg_reader.contigs[1].size_bytes)
        self.assertEqual([(100, 102)], ctg_reader.contigs[0].breakpoints)
        self.assertEqual([(50, 55), (250, 255)], ctg_reader.contigs[1].breakpoints)
        self.assertEqual(5, ctg_reader.contigs[0].avg_coverage)
        self.assertEqual(3, ctg_reader.contigs[1].avg_coverage)

    def test_read_three_features(self):
        ctg_reader = contig_reader.ContigReader(FEAT_DIR,
                                                [reader.feature_names[0],
                                                 reader.feature_names[1],
                                                 reader.feature_names[3]],
                                                process_count=1)
        
        result = ctg_reader.read_contigs(ctg_reader.contigs)

        # we read 2 contigs in total
        self.assertEqual(2, len(result))
        # each contig has 6 features (because 'ref_base' is one-hot encoded into 4 features)
        self.assertEqual(6, len(result[0]))
        self.assertEqual(6, len(result[1]))

        self.assertEqual(500, len(result[0]['ref_base_A']))
        self.assertIsNone(
            np.testing.assert_array_equal(np.array([1] * 498 + [0, 0]), result[0]['ref_base_A']))
        self.assertIsNone(
            np.testing.assert_array_equal(np.array([0] * 498 + [1, 1]), result[0]['ref_base_C']))
        self.assertFalse(np.any(result[0]['ref_base_G']))
        self.assertFalse(np.any(result[0]['ref_base_T']))
        coverage = result[0]['coverage']
        for pos in range(0, len(result[0]['ref_base_A'])):
            self.assertEqual(2 if 420 <= pos < 425 or pos < 5 else 0, coverage[pos])
            self.assertEqual(0.5 if 420 <= pos < 425 else 0, result[0]['num_query_C'][pos])

    if __name__ == '__main__':
        unittest.main()



import os
import bz2
import gzip
import shutil
import random
import logging
import multiprocessing as mp
from pkg_resources import resource_filename
from functools import partial
from distutils.spawn import find_executable
from subprocess import Popen, PIPE

import pysam

from resmico import utils


# functions
def _open(infile, mode='rb'):
    """Openning of input, regardless of compression."""
    if infile.endswith('.bz2'):
        return bz2.open(infile, mode)
    elif infile.endswith('.gz'):
        return gzip.open(infile, mode)
    else:
        return open(infile)


def _decode(x):
    """
    Decoding input, if needed
    """
    try:
        x = x.decode('utf-8')
    except AttributeError:
        pass
    return x


def file_exists(F, base_dir):
    """
    Does file exist as-is or at least with the appended base_dir?
    """
    if not os.path.exists(F):
        FF = os.path.join(base_dir, F)
        if not os.path.exists(FF):
            raise IOError(f'Cannot find file: "{F}" or "{FF}"')
        else:
            F = FF
    return F


def parse_input(infile):
    """Parsing tab-delimited input table."""
    base_dir = os.path.split(infile)[0]
    header = dict()
    idx = []
    with open(infile) as inF:
        for i, line in enumerate(inF):
            line = line.rstrip().split('\t')
            if line[0] == '':
                # blank line
                continue
            if len(line) < 4:
                raise ValueError('The input table must have >=4 columns!')
            if i == 0:
                # header
                header = {x.lower(): ii for ii, x in enumerate(line)}
                for k, v in {'Fasta': 'fasta', 'BAM': 'bam'}.items():
                    try:
                        _ = header[v]
                    except KeyError:
                        msg = f'Cannot find required column: "{k}"'
                        raise KeyError(msg)
                continue
            # body
            taxon = line[header['taxon']]
            sample = line[header['sample']]
            fasta = file_exists(line[header['fasta']], base_dir)
            bam = file_exists(line[header['bam']], base_dir)
            idx.append([bam, fasta, sample, taxon])
    return idx


def run_cmd(cmd):
    """
    Simple run of a command
    """
    cmd = [str(x) for x in cmd]
    logging.info('  CMD: {}'.format(' '.join(cmd)))
    p = Popen(cmd, stdout=PIPE, stderr=PIPE)
    output, err = p.communicate()
    rc = p.returncode
    if rc != 0:
        for line in output.decode().split('\n'):
            print(line)
        for line in err.decode().split('\n'):
            print(line)
        raise ValueError('Return code: {}'.format(rc))
    return rc


def uncomp_ref(fasta_file, tmpdir):
    """
    Uncompressing reference genome (if needed)
    """
    logging.info('  Uncompressing fasta...')
    outfile = os.path.join(tmpdir, os.path.split(fasta_file)[1])
    with _open(fasta_file) as inF, open(outfile, 'w') as outF:
        for line in inF:
            outF.write(_decode(line))
    return outfile


def faidx_ref(fasta_file, exe):
    """
    Indexing a reference via `samtools faidx`
    """
    logging.info('  Indexing fasta...')
    outfile = fasta_file + '.fai'
    cmd = [exe, 'faidx', fasta_file]
    run_cmd(cmd)
    return outfile


def get_basename(F):
    return os.path.splitext(os.path.split(F)[1])[0]


def index_bam(bam_file, exe, n_threads):
    """
    Indexing a bam file via `samtools index`
    """
    cmd = [exe, 'index', '-@', n_threads, bam_file]
    run_cmd(cmd)
    return bam_file + '.bai'


def sort_bam(bam_file, exe, tmpdir, n_threads):
    """
    Sorting a bam file via `samtools sort`
    """
    tmpdir = os.path.join(tmpdir, 'bam_sort_TMP')
    if not os.path.isdir(tmpdir):
        os.makedirs(tmpdir)
    outfile = os.path.splitext(bam_file)[0] + '_sorted.bam'
    cmd = [exe, 'sort', '-@', n_threads, '-o', outfile, bam_file]
    run_cmd(cmd)
    return outfile


def subsample_bam(bam_file, fasta_file, tmpdir, max_coverage, max_insert_size=30000):
    """
    Subsample the per-contig coverage from the bam file
    """
    bam = pysam.AlignmentFile(bam_file)
    fasta = pysam.FastaFile(fasta_file)
    outfile = os.path.join(tmpdir, get_basename(bam_file) + '_sub.bam')
    output = pysam.AlignmentFile(outfile, 'wb', template=bam)
    # getting contig lengths
    logging.info('  Subsampling reads...')
    contig_cov = {contig: 0 for contig in bam.references}
    for contig in bam.references:
        logging.info(f'    Processing contig: {contig}')
        contig_len = len(fasta.fetch(contig))
        for read in sorted(bam.fetch(contig), key=lambda k: random.random()):
            # new cov
            added_cov = read.query_length / contig_len
            # if max cov hit, skipping read
            if contig_cov[read.reference_name] + added_cov >= max_coverage:
                continue
            # if very large insert size, skipping
            if abs(read.template_length) > max_insert_size:
                continue
            # keeping read & tracking added coverage
            contig_cov[read.reference_name] = contig_cov[read.reference_name] + added_cov
            output.write(read)
        mean_cov = round(contig_cov[contig], 2)
        logging.info(f'    Final mean coverage: {mean_cov}')
    return outfile


def bam2feat(bam_file, fasta_file, outdir, exe, args):
    """
    Calling bam2feat on bam + fasta
    """
    cmd = [exe, '--procs', args.n_threads, '-queue_size', args.queue_size,
           '--window', args.window, '-breakpoint_margin', args.breakpoint_margin,
           '--o', outdir, '--bam_file', bam_file, '--fasta_file', fasta_file]
    run_cmd(cmd)
    return outdir


def _run_bam2feat(x, exe, outdir, tmpdir, args):
    """
    Per-BAM processing of input files and running bam2feat
    """
    logging.info(f'BAM file: {x[2]} => {x[0]}')
    logging.info(f'Fasta file: {x[3]} => {x[1]}')
    # tmpdir
    tmpdir = os.path.join(tmpdir, x[2], x[3])
    if not os.path.isdir(tmpdir):
        os.makedirs(tmpdir)
    logging.info(f'Temp Dir: {tmpdir}')
    # uncompress reference
    ref_tmp = uncomp_ref(x[1], tmpdir)
    # index ref
    ref_tmp_faidx = faidx_ref(ref_tmp, exe['samtools'])
    # index bam (if needed)
    bam_bai = x[0] + '.bai'
    if not os.path.isfile(bam_bai):
        bam_bai = index_bam(x[0], exe['samtools'], args.n_threads)
    # subsample bam
    bam_sub = subsample_bam(x[0], ref_tmp, tmpdir, args.max_coverage)
    # sort bam
    bam_sub = sort_bam(bam_sub, exe['samtools'], tmpdir, args.n_threads)
    # index bam
    bam_bai = index_bam(bam_sub, exe['samtools'], args.n_threads)
    # bam2feat
    if args.outdir_flat:
        outdir_feat = outdir
    else:
        outdir_feat = os.path.join(outdir, x[2], x[3])
    if not os.path.isdir(outdir_feat):
        os.makedirs(outdir_feat)
    logging.info(f'Outdir: {outdir_feat}')
    bam2feat(bam_sub, ref_tmp, outdir_feat, exe['bam2feat'], args)
    # clean up
    logging.info(f'Cleaning up Temp Dir: {tmpdir}')
    shutil.rmtree(tmpdir, ignore_errors=True)
    return [x[3], x[2], outdir_feat]


def run_bam2feat(bam_fasta, exe, args):
    """
    Main pipeline interface for bam2feat, including input file processing
    """
    # seed
    random.seed(args.seed)
    # outdir
    if not os.path.isdir(args.outdir):
        os.makedirs(args.outdir)
    # per-BAM processing
    func = partial(_run_bam2feat, exe=exe,
                   outdir=args.outdir,
                   tmpdir=args.tmpdir,
                   args=args)
    if args.n_proc < 2:
        logging.info('Processing input...')
        res = map(func, bam_fasta)
    else:
        logging.info('Processing input in parallel...')
        set_logger(logging.WARNING)
        pool = mp.Pool(args.n_proc)
        res = pool.map(func, bam_fasta)
    return [x for x in res]


def write_feat_table(feat_files, outdir):
    """
    Writing tab-delim feature file table 
    """
    outfile = os.path.join(outdir, 'feature_files.tsv')
    with open(outfile, 'w') as outF:
        header = ['Taxon', 'Sample', 'feature_file']
        outF.write('\t'.join(header) + '\n')
        for line in feat_files:
            outF.write('\t'.join(line) + '\n')
    logging.info(f'Feature file table written: {outfile}')


def main(args):
    logging.info('Starting bam2feat...')
    # check that bam2feat is available
    exe = {'bam2feat': resource_filename('resmico', 'bam2feat'),
           'samtools': 'samtools'}
    for k, v in exe.items():
        if find_executable(v) is None:
            raise OSError(f'Cannot find executable: {k}')
    # parse input table
    logging.info(f'Parsing input table: {args.input_table}')
    bam_fasta = parse_input(args.input_table)
    # run bam2feat in parallel
    feat_files = run_bam2feat(bam_fasta, exe, args)
    # writing table
    write_feat_table(feat_files, args.outdir)
    # clean up
    shutil.rmtree(args.tmpdir, ignore_errors=True)


if __name__ == '__main__':
    pass



import argparse
import logging
from resmico import bam2feat


# functions
def get_desc():
    desc = 'Convert >=1 contig fasta + mapped-reads BAM to a feature table'
    return desc


def parse_args(test_args=None, subparsers=None):
    desc = get_desc()
    epi = """DESCRIPTION:
    Convert 1 or more contig fasta file and associated mapped paired-end
    reads (BAM file) to one or more 1 resmico feature table.
    The input_table maps the fasta and BAM files.
    The defaults are the same as used to generate all training/test data in the
    Mineeva et al., 2022 manuscript.
    --n-proc sets the per-BAM parallelization.
    --n-threads sets the per-command (eg., samtools) parallelization.

    Note: the input bam file(s) must be sorted
    """
    if subparsers:
        parser = subparsers.add_parser('bam2feat', description=desc, epilog=epi,
                                       formatter_class=argparse.RawTextHelpFormatter)
    else:
        parser = argparse.ArgumentParser(description=desc, epilog=epi,
                                         formatter_class=argparse.RawTextHelpFormatter)
    # args
    parser.add_argument('input_table', type=str,
                        help='A tab-delim table with the columns: Taxon,Fasta,Sample,BAM.\n'
                             'The columns can be in any order; capitalization does not matter.\n'
                             'Taxon: name associated with the fasta file\n'
                             'Sample: name associated with the sorted BAM file\n')
    parser.add_argument('--outdir', default='resmico-bam2feat', type=str,
                        help='Output base directory (default: %(default)s)')
    parser.add_argument('--outdir-flat', action='store_true', default=False,
                        help='No nested directories for output files?'
                             'This is useful if just processing 1 sample (default: %(default)s)')
    parser.add_argument('--tmpdir', default='resmico-bam2feat_TMP', type=str,
                        help='Temporary file directory (default: %(default)s)')
    parser.add_argument('--max-coverage', default=20.0, type=float,
                        help='Subsample mapped reads to this max coverage per-contig'
                             ' (default: %(default)s)')
    parser.add_argument('--window', default=6, type=int,
                        help='Sliding window size for sequence entropy & GC content'
                             ' (default: %(default)s)')
    parser.add_argument('--breakpoint-margin', default=50, type=int,
                        help='Maximum offset (to left or right) around the breaking point\n'
                             'used when creating a chunk (default: %(default)s)')
    parser.add_argument('--queue-size', default=32, type=int,
                        help='Maximum size of the queue for stats waiting to be written to disk,\n'
                             'before blocking (default: %(default)s)')
    parser.add_argument('--seed', default=8192, type=int,
                        help='Seed for reproducible subsampling (default: %(default)s)')
    parser.add_argument('--n-proc', default=1, type=int,
                        help='No. of sorted BAM files to process in parallel (default: %(default)s)')
    parser.add_argument('--n-threads', default=1, type=int,
                        help='No. threads to pass to samtools & bam2feat (default: %(default)s)')

    # test args
    if test_args:
        args = parser.parse_args(test_args)
        return args
    # return
    return parser


def main(args=None):
    logging.basicConfig(format='%(asctime)s - %(message)s', level=logging.DEBUG)
    # Input
    if args is None:
        args = parse_args()
        print()
        print(args)
        print()
    # Main interface
    bam2feat.main(args)


# main
if __name__ == '__main__':
    pass



import argparse
from pkg_resources import resource_filename

def add_common_args(parser: argparse.ArgumentParser):
    """
    Adds arguments common to both training and evaluation to parser.
    """
    # default stats for n9k-train training dataset
    pkg_stats = resource_filename('resmico', 'model/stats_cov.json')
    parser_g1 = parser.add_argument_group('General arguments')
    parser_g1.add_argument('--feature-files-path', default='.', type=str,
                           help='Path to the feature files produced by ResMiCo-SM (default: %(default)s).\n'
                           '2 options are available:\n'
                           '  1) Provide the base path, and subdirectories will be searched.\n'
                           '  2) Provide a file that lists all stats files.\n'
                           '     Note: associated files (eg., toc files) must be in the same directories.\n')
    parser_g1.add_argument('--feature-file-match', default='', type=str,
                           help='String that paths to feature files must match.\n'
                           'Example: use "0.005" to select file paths containing "0.005".\n'
                           'Use "" to match all paths (default: %(default)s)')
    parser_g1.add_argument('--stats-file', default=pkg_stats,
                           help='File containing the feature means/stdevs of the training set.\n'
                           'Set to an empty string when training on a new data set. \n(default: %(default)s)')
    parser_g1.add_argument('--save-path', default='.', type=str,
                           help='Directory where to save output (default: %(default)s)')
    parser_g1.add_argument('--save-name', default='resmico', type=str,
                           help='Prefix for name in the save_path (default: %(default)s)')
    parser_g1.add_argument('--n-procs', default=1, type=int,
                           help='Number of parallel processes (default: %(default)s)')
    parser_g1.add_argument('--gpu-eval-mem-gb', default=1.0, type=float,
                           help='Amount of GPU memory used for validation data.\n'
                           'The amount will be divided per GPU (default: %(default)s)')
    parser_g1.add_argument('--max-len', default=20000, type=int,
                           help='Max contig length; larger contigs are split (default: %(default)s)')
    parser_g1.add_argument('--min-contig-len', default=1000, type=int,
                           help='Ignore contigs with length smaller than this value (default: %(default)s)')
    parser_g1.add_argument('--min-avg-coverage', default=1.0, type=float,
                           help='Minimum average coverage for a contig to be considered during evaluation\n'
                           'or training (default: %(default)s)')
    parser_g1.add_argument('--features', nargs='+', help='Features used by resmico (default: %(default)s)',
                           default=[
                               'num_query_A',
                               'num_query_C', 
                               'num_query_G', 
                               'num_query_T',
                               'mean_mapq_Match',
                               'stdev_al_score_Match',
                               'mean_al_score_Match',
                               'mean_insert_size_Match',
                               'coverage',
                               'min_al_score_Match',
                               'num_SNPs',
                               'min_insert_size_Match',
                               'num_proper_Match',
                               'num_orphans_Match'])
    parser_g1.add_argument('--val-ind-f', default=None, type=str,
                           help='Validation data indices (default: %(default)s)')
    parser_g1.add_argument('--seed', default=12, type=int,
                           help='Seed used for numpy.random and tf (default: %(default)s)')
    parser_g1.add_argument('--log-level', default='INFO',
                           choices = ['CRITICAL', 'FATAL', 'ERROR', 'WARNING', 'INFO', 'DEBUG'],
                           help='Logging level (default: %(default)s)')
    parser_g1.add_argument('--no-cython', dest='no_cython', action='store_true',
                           help='If set, data is read using pure Python rather than using the Cython bindings\n'
                           '(about 2x slower; only useful for debugging)')



#!/usr/bin/env python
from __future__ import print_function
import sys,os
import glob
import gzip
import argparse
import logging
import random

desc = 'Copying data from ResMiCo-SM test to DL-test data directory'
epi = """DESCRIPTION:
Simple script to convert the files in the features/
directory created by ResMiCo-SM to data files for
running the ResMiCo unit tests.

In order to make sure that there are enough misassemblies
in the feature tables for training, the miassemblies are
randomly assigned to contigs (random.randint(0,1))
"""
parser = argparse.ArgumentParser(description=desc,
                                 epilog=epi,
                                 formatter_class=argparse.RawTextHelpFormatter)
parser.add_argument('features_dir', metavar='features_dir', type=str,
                    help='ResMiCo-SM features direcotry')
parser.add_argument('out_dir', metavar='out_dir', type=str,
                    help='Output directory')
parser.add_argument('-l', '--lines', type=int, default=200000,
                    help='No. of lines from each feature table to use (default: %(default)s)')
parser.add_argument('--version', action='version', version='0.0.1')

logging.basicConfig(format='%(asctime)s - %(message)s', level=logging.DEBUG)


def main(args):
    if not os.path.isdir(args.out_dir):
        os.makedirs(args.out_dir)
        
    # formatting feature files table
    feat_file_tbl_file = os.path.join(args.features_dir, 'feature_files.tsv')
    if not os.path.isfile(feat_file_tbl_file):
        raise IOError('Cannot find {}'.format(feat_file_tbl_file))
    feat_file_tbl_out = os.path.join(args.out_dir, 'feature_files.tsv')
    with open(feat_file_tbl_file) as inF, open(feat_file_tbl_out,'w') as outF:
        header = {}
        for i,line in enumerate(inF):
            line = line.rstrip().split('\t')
            if i == 0:
                header = {x:i for i,x in enumerate(line)}         
            if i > 0:
                ii = header['feature_file']
                line[ii] = '/'.join(line[ii].split('/')[3:])
            outF.write('\t'.join(line) + '\n')
    logging.info('Feature table file written: {}'.format(feat_file_tbl_out))

        
    # formatting feature files
    P = os.path.join(args.features_dir, '*', '*', '*', '*', 'features.tsv.gz')
    feature_files = glob.glob(P)
    if len(feature_files) == 0:
        raise IOError('Cannot find feature files at {}'.format(P))
    for feat_file in feature_files:
        logging.info('Processing feature file: {}'.format(feat_file))
        out_file = os.path.join(args.out_dir, '/'.join(feat_file.split('/')[-5:]))
        out_dir = os.path.split(out_file)[0]
        if not os.path.isdir(out_dir):
            os.makedirs(out_dir)
        with gzip.open(feat_file, 'rb') as inF, gzip.open(out_file, 'wt') as outF:
            cur_contig = None
            last_contig = ''
            header = {}
            misasmbl = str(random.randint(0,1))   # random assignment of misassembly
            for i,line in enumerate(inF):
                line = line.decode('utf8').rstrip().split('\t')
                if i == 0:
                    header = {x:i for i,x in enumerate(line)}
                else:
                    cur_contig = line[header['contig']]
                    if cur_contig != last_contig:
                        misasmbl = str(random.randint(0,1))   # random assignment of misassembly                    
                    line[header['Extensive_misassembly']] = misasmbl
                line = '\t'.join(line) + '\n'
                outF.write(line)
                last_contig = cur_contig
                if i + 2 >= args.lines:
                    break                                               
        logging.info('Feature file written: {}'.format(out_file))


if __name__ == '__main__':
    args = parser.parse_args()
    main(args)



# define NPY_NO_DEPRECATED_API NPY_1_7_API_VERSION
from setuptools import Extension, setup
from Cython.Build import cythonize
import numpy
import platform

compile_args = ['-std=c++11']
link_args = ['-std=c++11']
if platform.system() != 'Darwin':
    compile_args.append('-fopenmp')
    link_args.append('-fopenmp')

sourcefiles = ['reader.pyx', 'contig_reader.cpp']

extensions = [Extension('reader',
                        sourcefiles,
                        language="c++",
                        extra_compile_args=compile_args,
                        extra_link_args=link_args,
                        libraries=['z'],
                        )]

setup(
    ext_modules=cythonize(extensions,
                          include_path=['.', numpy.get_include()],
                          annotate=True,
                          compiler_directives={'language_level': "3"}
                          ),
    include_dirs=['.', numpy.get_include()],
)



import os
import logging
import math
import time
import atexit

import numpy as np
import pandas as pd
import tensorflow as tf
from tensorflow.keras.models import load_model
from tensorflow.keras import backend as K
from sklearn.metrics import recall_score, average_precision_score

from resmico import contig_reader
from resmico import models_fl as Models
from resmico import utils


def main(args):
    """
    Trains ResMiCo on binary data files produced by ResMiCo-SM.
    """
    np.random.seed(args.seed)
    if not os.path.exists(args.save_path):
        os.makedirs(args.save_path)

    # disable tf allocating all memory on the device at the very beginning, as this seems to be causing
    # LSTM's to crash at prediction time.
    gpu_devices = tf.config.experimental.list_physical_devices("GPU")
    for device in gpu_devices:
        tf.config.experimental.set_memory_growth(device, True)

    logging.info('Building Tensorflow model...')
    logging.info(args)
    strategy = tf.distribute.MirroredStrategy()

    with strategy.scope():
        resmico = Models.Resmico(args)
        if os.path.exists(args.model_checkpoint):
            logging.info(f'Loading model: {args.model_checkpoint}')
#             custom_obj = {'class_recall_0': utils.class_recall_0, 'class_recall_1': utils.class_recall_1,
#                   'GlobalMaskedMaxPooling1D': Models.GlobalMaskedMaxPooling1D}
            resmico.net.load_weights(args.model_checkpoint)
            logging.info('Model loaded')
            
    resmico.print_summary()

    # tensorboard logs
    tb_logs = tf.keras.callbacks.TensorBoard(log_dir=os.path.join(args.save_path, args.save_name),
                                             histogram_freq=0,
                                             write_graph=True,
                                             write_images=True)

    logging.info('Loading contig data...')
    reader = contig_reader.ContigReader(args.feature_files_path, args.features, args.n_procs,
                                        args.no_cython, args.stats_file, min_len=args.min_contig_len,
                                        min_avg_coverage=args.min_avg_coverage,
                                        feature_file_match=args.feature_file_match)

    # separate data into 90% for training and 10% for evaluation
    all_idx = np.arange(len(reader))
    if args.val_ind_f:
        logging.info(f'Split data: using {args.val_ind_f} for validation, for training everything else')
        eval_idx = list(pd.read_csv(args.val_ind_f)['val_ind'])
        train_idx = list(set(all_idx) - set(eval_idx))
    else:
        np.random.shuffle(all_idx)
        train_idx = all_idx[:(9 * len(reader)) // 10]
        eval_idx = all_idx[(9 * len(reader)) // 10:]
        df = pd.DataFrame(eval_idx, columns=['val_ind'])
        if os.path.isfile(args.feature_files_path):
            fname = os.path.join(os.path.split(args.feature_files_path)[0], "evaluation_indices.csv")
        else:
            fname = os.path.join(args.feature_files_path.split(",")[0], "evaluation_indices.csv")
        df.to_csv(fname)
        logging.info(f'Evaluation indices saved to: {fname}')
    logging.info(f'Using {len(train_idx)} contigs for training, {len(eval_idx)} contigs for evaluation')

    # create data generators for training data and evaluation data
    train_data = Models.BinaryDatasetTrain(reader, train_idx, args.batch_size, args.features, args.max_len,
                                           args.num_translations, args.max_translation_bases, args.fraq_neg,
                                           args.cache_train or args.cache, args.log_progress, resmico.convoluted_size,
                                           resmico.fixed_length, args.weight_factor)
    # convert the slow Keras train_data of type Sequence to a tf.data object
    # first, we convert the keras sequence into a generator-like object
    data_iter = lambda: (s for s in train_data)

    # second, we use tf.data.Dataset.from_generator to create a tf.data.Dataset object and use this for training
    train_data_tf = tf.data.Dataset.from_generator(
        data_iter,
        output_signature=(
            (tf.TensorSpec(shape=(args.batch_size, None, len(train_data.expanded_feature_names)), dtype=tf.float32),
             tf.TensorSpec(shape=(args.batch_size, None), dtype=tf.bool)),
            tf.TensorSpec(shape=(args.batch_size), dtype=tf.uint8),
            tf.TensorSpec(shape=(args.batch_size), dtype=tf.float32)))

    # add a prefetch option that builds the next batch ready for consumption by the GPU as it is working on
    # the current batch.
    train_data_tf = train_data_tf.prefetch(2 * strategy.num_replicas_in_sync)

    # set the sharding policy to DATA in order to avoid Tensorflow ugly console barf
    options = tf.data.Options()
    options.experimental_distribute.auto_shard_policy = tf.data.experimental.AutoShardPolicy.DATA
    train_data_tf = train_data_tf.with_options(options)

    np.seterr(all='raise')
    eval_data = Models.BinaryDatasetEval(reader, eval_idx, args.features, args.max_len, args.max_len-500,
                                         int(args.gpu_eval_mem_gb * 1e9 * 0.8), args.cache_validation or args.cache,
                                         args.log_progress, resmico.convoluted_size, 
                                         resmico.fixed_length, batch_size=args.batch_size)

    eval_data_y = np.array([0 if reader.contigs[idx].misassembly == 0 else 1 for idx in eval_data.indices])

    # convert the slow Keras eval_data of type Sequence to a tf.data object
    data_iter = lambda: (s for s in eval_data)
    eval_data_tf = tf.data.Dataset.from_generator(
        data_iter,
        output_signature=(
            # first dimension is batch size, second is contig length, third is number of features
            (tf.TensorSpec(shape=(None, None, len(eval_data.expanded_feature_names)), dtype=tf.float32),
             # first dimension is batch size, second is contig length (no third dimension,
             # as all features are masked the same way)
             tf.TensorSpec(shape=(None, None), dtype=tf.bool)),
            tf.TensorSpec(shape=(None), dtype=tf.bool)
        ))
    eval_data_tf = eval_data_tf.prefetch(4 * strategy.num_replicas_in_sync)
    eval_data_tf = eval_data_tf.with_options(options)  # avoids Tensorflow ugly console barf

    logging.info('Training network...')
    num_epochs = 2
    train_data_tf = train_data_tf.repeat(num_epochs)
    auc_val_best = 0
    auc_val_prev = 0
    best_file = None
    for epoch in range(math.ceil(args.n_epochs / num_epochs)):
        start = time.time()
        resmico.net.fit(x=train_data_tf,
                        epochs=num_epochs,
                        steps_per_epoch=len(train_data),
                        workers=args.n_procs,
                        use_multiprocessing=True,
                        max_queue_size=max(args.n_procs, 10),
#                         callbacks=[tb_logs],
                        verbose=2)

        duration = time.time() - start
        logging.info(f'Fitted {num_epochs} epochs in {duration:.0f}s')
        train_data.on_epoch_end()

        logging.info('Starting validation')
        start = time.time()
        eval_data_flat_y = resmico.predict(x=eval_data_tf,
                                           workers=args.n_procs,
                                           use_multiprocessing=True,
                                           max_queue_size=max(args.n_procs, 10),
                                           verbose=2)
        eval_data_predicted_y = eval_data.group(eval_data_flat_y)

        auc_val = average_precision_score(eval_data_y, eval_data_predicted_y)
        recall1_val = recall_score(eval_data_y, eval_data_predicted_y > 0.5, pos_label=1)
        recall0_val = recall_score(eval_data_y, eval_data_predicted_y > 0.5, pos_label=0)
        logging.info(f'Validation scores after {(epoch + 1) * num_epochs} epochs: aucPR: {auc_val} - '
                     f'recall1: {recall1_val} - recall0: {recall0_val} - mean: {np.mean(eval_data_predicted_y)}')
        duration = time.time() - start
        logging.info(f'Validation done in {duration:.0f}s')

        # update the learning rate
        cur_epoch = (epoch + 1) * num_epochs
        if cur_epoch > 10 and auc_val < auc_val_prev:
            lr_old = K.get_value(resmico.net.optimizer.lr)
            K.set_value(resmico.net.optimizer.lr, lr_old * 0.8)  # changed for 120h jobs
            logging.info(f'Updated learning rate from: {lr_old} to {K.get_value(resmico.net.optimizer.lr)}')

        auc_val_prev = auc_val
        if auc_val > auc_val_best:
            if best_file:  # delete old best model
                try:
                    os.remove(best_file)
                except OSError:
                    logging.warning('Unable to remove: ' + best_file)
            auc_val_best = auc_val
            best_file = os.path.join(args.save_path, '_'.join(
                ['mc_epoch', str(cur_epoch), 'aucPR', str(auc_val_best)[:5], args.save_name,
                 'model.h5']))
            resmico.save(best_file)
            logging.info(f'New best model written to: {best_file}')

    # saving
    logging.info('Saving trained model...')
    outfile = os.path.join(args.save_path, args.save_name + '_' + str(auc_val)[:5] + '_e' + str(args.n_epochs) + '.h5')
    resmico.save(outfile)
    logging.info(f'Latest model written to: {outfile}')
    
    # exit
    atexit.register(strategy._extended._collective_ops._pool.close)


if __name__ == '__main__':
    pass



import os
import pytest
import logging

# test/data dir
test_dir = os.path.join(os.path.dirname(__file__))
data_dir = os.path.join(test_dir, 'data')

# tests
def test_helps(script_runner):
    ret = script_runner.run('resmico', 'filter', '-h')
    assert ret.success, ret.print()

def test_filter(tmpdir, script_runner):
    save_path = tmpdir.mkdir('save_dir')
    pred_file = os.path.join(data_dir, 'UHGG-n9', 'resmico_predictions.csv')
    contig_file = os.path.join(data_dir, 'UHGG-n9', 'GUT_GENOME031704.fna.gz')
    ret = script_runner.run('resmico', 'filter',
                            '--score-cutoff', '0.02',
                            '--outdir', str(save_path),
                            pred_file, contig_file)
    assert ret.success, ret.print()

def test_filter_list(tmpdir, script_runner):
    save_path = tmpdir.mkdir('save_dir')
    pred_file = os.path.join(data_dir, 'UHGG-n9', 'resmico_predictions.csv')
    contig_files = os.path.join(data_dir, 'UHGG-n9', 'contig_files.txt')
    ret = script_runner.run('resmico', 'filter',
                            '--score-cutoff', '0.02',
                            '--outdir', str(save_path),
                            pred_file, contig_files)
    assert ret.success, ret.print()

def test_filter_outfile(tmpdir, script_runner):
    save_path = tmpdir.mkdir('save_dir')
    pred_file = os.path.join(data_dir, 'UHGG-n9', 'resmico_predictions.csv')
    contig_files = os.path.join(data_dir, 'UHGG-n9', 'contig_files.txt')
    ret = script_runner.run('resmico', 'filter',
                            '--score-cutoff', '0.02',
                            '--outdir', str(save_path),
                            '--outfile', 'contigs-filtered.fna',
                            pred_file, contig_files)
    assert ret.success, ret.print()



# distutils: language = c++
cimport cython
from libc.stdlib cimport malloc, free

from array import array
from libc.stdint cimport uint32_t
from libc.stdint cimport uint64_t
from libc.stdint cimport uint16_t
from libc.stdint cimport uint8_t

import numpy as np
from cython.parallel import parallel, prange
from cpython.mem cimport PyMem_Malloc, PyMem_Free

from typing import List

cdef extern from "Python.h":
    char *PyBytes_AS_STRING(object)

# define the available feature, their size in the feature file, their size as numpy arrays, and if they need to
# be normalized
DEF N_FEATURES = 25
feature_tuples = [('ref_base', np.uint8, np.float32, False),  # because we use one-hot encoding, so 4 bytes
                  ('coverage', np.uint16, np.float32, False),  # converting to float32 bc it's going to be normalized
                  ('num_query_A', np.uint16, np.float32, False),
                  ('num_query_C', np.uint16, np.float32, False),
                  ('num_query_G', np.uint16, np.float32, False),
                  ('num_query_T', np.uint16, np.float32, False),
                  ('num_SNPs', np.uint16, np.float32, False),
                  ('num_discordant', np.uint16, np.float32, False),
                  ('min_insert_size_Match', np.uint16, np.float32, True),
                  ('mean_insert_size_Match', np.float32, np.float32, True),
                  ('stdev_insert_size_Match', np.float32, np.float32, True),
                  ('max_insert_size_Match', np.uint16, np.float32, True),
                  ('min_mapq_Match', np.uint8, np.float32, True),
                  ('mean_mapq_Match', np.float32, np.float32, True),
                  ('stdev_mapq_Match', np.float32, np.float32, True),
                  ('max_mapq_Match', np.uint8, np.float32, True),
                  ('min_al_score_Match', np.int8, np.float32, True),
                  ('mean_al_score_Match', np.float32, np.float32, True),
                  ('stdev_al_score_Match', np.float32, np.float32, True),
                  ('max_al_score_Match', np.int8, np.float32, True),
                  ('num_proper_Match', np.uint16, np.float32, False),
                  ('num_orphans_Match', np.uint16, np.float32, False),
                  ('num_proper_SNP', np.uint16, np.float32, False),
                  ('seq_window_perc_gc', np.float32, np.float32, True),
                  ('seq_window_entropy', np.float32, np.float32, True),
                   ]

float_feature_names = [f[0] for f in feature_tuples if f[3]]

feature_names = [f[0] for f in feature_tuples]
feature_types = [f[1] for f in feature_tuples]
feature_np_types = [f[2] for f in feature_tuples]
feature_sizes = [np.dtype(feature_types[i]).itemsize for i in range(N_FEATURES)]
feature_np_sizes = [np.dtype(feature_types[i]).itemsize for i in range(N_FEATURES)]

bytes_per_base = sum(feature_sizes)

assert len(feature_names) == N_FEATURES
assert len(feature_sizes) == N_FEATURES
assert len(feature_types) == N_FEATURES

cdef extern from 'contig_reader.hpp':
    cdef void read_contig_features(const char *fname, uint64_t offset, uint32_t size_bytes,
                          uint32_t length_bases, uint32_t num_features,
                          uint16_t b_per_base, uint8_t *feature_mask,
                          uint8_t *feature_sizes_bytes, char **features) nogil
    cdef void read_contig_features_buf(const char *fname, uint64_t offset, uint32_t size_bytes,
                                   uint32_t length_bases, uint32_t num_features,
                                   uint16_t b_per_base, uint8_t *feature_mask,
                                   uint8_t *feature_sizes_bytes, char *buf, char ** features, int thread) nogil


@cython.boundscheck(False)
cdef read_contig_cpp(const char* file_name, uint32_t length, uint64_t offset, uint32_t size, uint8_t[:] feature_mask):
    cdef uint32_t[2] lengths = {1, length}
    cdef char[:] view
    np_data = [None] * N_FEATURES
    cdef char *all_data[25]
    for i in range(N_FEATURES):
        np_data[i] = np.empty([lengths[feature_mask[i]]], dtype = feature_types[i])
        view = np_data[i].view(np.int8)
        all_data[i] = &view[0]

    cdef uint8_t feature_sizes_bytes[N_FEATURES]
    feature_sizes_bytes[:] = feature_sizes

    read_contig_features(file_name, offset, size, length, N_FEATURES, bytes_per_base, &feature_mask[0],
                         &feature_sizes_bytes[0], &all_data[0])

    result = {feature_name: data for feature_name, data in zip(feature_names, np_data)}
    return result


def read_contig_py(str file_name, int length, int offset, int size, py_feature_names):
    py_feature_mask = [1 if feature in py_feature_names else 0 for feature in feature_names]
    cdef uint8_t[:] feature_mask = np.array(py_feature_mask, dtype=np.uint8)
    result = read_contig_cpp(file_name.encode('utf-8'), length, offset, size, feature_mask)
    return {key: result[key] for key in py_feature_names}

# Reads contig features from #file_names and returns a list of {'feature_name', 'feature_data'} dictionaries, for each
# contig. Data is read in parallel (using Cython bindings).
# Parameters:
#   file_names: names of the binary files to read data from, one per contig; file names are not necessarily distinct,
#           since each file contains data for many (hundreds) of contigs
#   py_lengths: the length of each contig
#   py_offsets: the position in the file where the contig data begins
#   py_sizes: the size of data in bytes, for each contig (used to allocate memory in the C code)
#   py_feature_mask: 0/1 mask denoting the features that need to be read
#   num_threads: how many threads to use to read the data
@cython.boundscheck(False)
@cython.wraparound(False)
def read_contigs_py(file_names:List[bytes], py_lengths: List[int],  py_offsets: List[int],  py_sizes: List[int],
                    py_feature_mask: List[int], int num_threads):
    assert len(file_names) == len(py_lengths) == len(py_offsets) == len(py_sizes)
    cdef uint32_t contig_count = len(file_names)
    cdef int max_len = max(py_lengths)
    cdef int[:] lengths = array('i', py_lengths)
    cdef uint64_t[:] offsets = array('L', py_offsets)
    cdef int[:] sizes = array('i', py_sizes)
    cdef uint8_t[:] feature_mask = array('B', py_feature_mask)
    # the buffer used by the C++ code to write the features into; one buffer per contig; for each contig one buffer
    # per feature
    cdef char ***all_data = <char ***> PyMem_Malloc(sizeof(char **) * contig_count)

    py_all_data = [None] * contig_count
    cdef uint32_t[2] arr_len
    cdef char[:] view
    views = [] # keep all views in a list to avoid garbage collection
    for ctg_idx in range(contig_count):
        np_data = [None] * N_FEATURES
        np_data_int8 = [None] * N_FEATURES
        all_data[ctg_idx] = <char **> PyMem_Malloc(sizeof(char **) * N_FEATURES)
        for feat_idx in range(N_FEATURES):
            if feature_mask[feat_idx]:
                np_data[feat_idx] = np.empty(lengths[ctg_idx], dtype=feature_types[feat_idx])
                view = np_data[feat_idx].view(np.int8)
                views.append(view)
                all_data[ctg_idx][feat_idx] = &view[0]
            else:
                all_data[ctg_idx][feat_idx] = NULL
        py_all_data[ctg_idx] = np_data

    cdef uint8_t feature_sizes_bytes[N_FEATURES]
    feature_sizes_bytes[:] = feature_sizes


    cdef char ** c_file_names = <char **>PyMem_Malloc(sizeof(char*) * contig_count)
    for i in range(contig_count):
        c_file_names[i] = PyBytes_AS_STRING(file_names[i])

    # This is the code that is actually parallelized
    cdef Py_ssize_t ctg_idx_c
    cdef uint32_t bytes_per_base_c = bytes_per_base
    cdef char * buf
    with nogil, parallel(num_threads = num_threads):
        # the buffer used by the C++ code to unzip the data (one buffer for each thread)
        buf = <char *> malloc(sizeof(char) * max_len * bytes_per_base_c + 4)
        for ctg_idx_c in prange(contig_count, schedule='guided'):
            read_contig_features_buf(c_file_names[ctg_idx_c], offsets[ctg_idx_c], sizes[ctg_idx_c], lengths[ctg_idx_c],
                                 N_FEATURES, bytes_per_base_c, &feature_mask[0], &feature_sizes_bytes[0],
                                 buf, all_data[ctg_idx_c], cython.parallel.threadid())
        free(buf)

    results = []
    for ctg_idx in range(contig_count):
        results.append({feature_name: data for feature_name, data in zip(feature_names, py_all_data[ctg_idx])
                        if data is not None })

    for feat_idx in range(contig_count):
        PyMem_Free(all_data[feat_idx])
    PyMem_Free(all_data)
    PyMem_Free(c_file_names)
    return results



#!/usr/bin/env python
import sys
import argparse

from resmico.commands import bam2feat
from resmico.commands import train
from resmico.commands import evaluate
from resmico.commands import filter_contigs


def main(args=None):
    """Main entry point for application
    """
    if args is None:
        args = sys.argv[1:]

    desc = 'ResMiCo: mis-assembly detection with deep learning'
    epi = """DESCRIPTION:
    Usage: resmico <subcommand> <subcommand_params>
    subcommand is one of: bam2feat, train, evaluate, filter
    Example: resmico train -h

    For general info, see https://github.com/leylabmpi/resmico/
    """
    parser = argparse.ArgumentParser(description=desc,
                                     epilog=epi,
                                     formatter_class=argparse.RawTextHelpFormatter)

    # subparsers
    subparsers = parser.add_subparsers()
    # bam2feat
    b2f = bam2feat.parse_args(subparsers=subparsers)
    b2f.set_defaults(func=bam2feat.main)
    # train
    trainer = train.parse_args(subparsers=subparsers)
    trainer.set_defaults(func=train.main)
    # evaluate
    evaluator = evaluate.parse_args(subparsers=subparsers)
    evaluator.set_defaults(func=evaluate.main)
    # filter
    filter_cntgs = filter_contigs.parse_args(subparsers=subparsers)
    filter_cntgs.set_defaults(func=filter_contigs.main)    

    # parsing args
    if args:
        args = parser.parse_args(args)
    else:
        args = parser.parse_args()

    # running subcommands
    if len(vars(args)) > 0:
        args.func(args)
    else:
        parser.parse_args(['--help'])


if __name__ == '__main__':
    main()



import os
import pytest
import logging

# test/data dir
test_dir = os.path.join(os.path.dirname(__file__))
data_dir = os.path.join(test_dir, 'data')

# tests
def test_helps(script_runner):
    ret = script_runner.run('resmico', 'bam2feat', '-h')
    assert ret.success, ret.print()




import os
import pytest
import logging

# test/data dir
test_dir = os.path.join(os.path.dirname(__file__))
data_dir = os.path.join(test_dir, 'data')

# tests
def test_helps(script_runner):
    ret = script_runner.run('resmico', 'train', '-h')
    assert ret.success

def test_train(tmpdir, script_runner):
    save_path = tmpdir.mkdir('save_dir')
    input_path = os.path.join(data_dir, 'n10', 'features')
    ret = script_runner.run('resmico', 'train',
                            '--n-epochs', '2',            
                            '--save-path', str(save_path),
                            '--feature-files-path', input_path)
    assert ret.success, ret.print()

def test_train_filelist(tmpdir, script_runner):
    save_path = tmpdir.mkdir('save_dir')
    input_path = os.path.join(data_dir, 'n10',
                              'features', 'file_list.txt')
    ret = script_runner.run('resmico', 'train',
                            '--n-epochs', '2',            
                            '--save-path', str(save_path),
                            '--feature-files-path', input_path)
    assert ret.success, ret.print()



from __future__ import print_function
from pkg_resources import resource_filename
import os
import argparse
import logging
from resmico import filter_contigs

# functions
def get_desc():
    desc = 'Filter out misassembled contigs'
    return desc

def parse_args(test_args=None, subparsers=None):
    desc = get_desc()
    epi = """DESCRIPTION:
    Filter out contigs predicted to be misassembled based on the resmico prediction score.
    Contigs must be fasta-formatted, with contig names matching those in the prediction score table.
    Filtered contig fasta files will be written to the designated output directory.
    Make sure to adjust the --score-cutoff as needed!
    """
    if subparsers:
        parser = subparsers.add_parser('filter', description=desc, epilog=epi,
                                       formatter_class=argparse.RawTextHelpFormatter)
    else:
        parser = argparse.ArgumentParser(description=desc, epilog=epi,
                                         formatter_class=argparse.RawTextHelpFormatter)
    # args
    parser.add_argument('prediction_table',  metavar='prediction_table', type=str, 
                        help='A table that includes per-contig misassembly predictions. '
                             'The table should be formatted as the output from "resmico predict"')
    parser.add_argument('fasta',  metavar='fasta', type=str, nargs='+',
                        help='>=1 fasta file containing the associated contigs.\n'
                        'Alternatively, a list of fasta files (1 per line) can be provided.\n'
                        'The contig names must match those in the prediction_table')
    parser.add_argument('--outdir', default='resmico-filter', type=str, 
                        help='Output directory (default: %(default)s)')
    parser.add_argument('--outfile', default=None, type=str, 
                        help='Output file name. If None, then the input file name will be used (default: %(default)s)')
    parser.add_argument('--score-cutoff', default=0.8, type=float, 
                        help='Prediction score cutoff for filtering: >=[score] will be filtered'
                        ' (default: %(default)s)')
    parser.add_argument('--score-delim', default=',', type=str, 
                        help='Delimiter for predictions table (default: %(default)s)')
    parser.add_argument('--add-score', action='store_true', default=False,
                        help='Add prediction score to sequence header "score=<score>"?'
                        ' (default: %(default)s)')
    parser.add_argument('--error-on-missing', action='store_true', default=False,
                        help='Error if a contig does not have a prediction? (default: %(default)s)')
    parser.add_argument('--min-length', default=0, type=int, 
                        help='Only apply filtering to contigs > min-length.\n'
                             'Shorter contigs are retained, regardless of the prediction score.\n'
                             'If <1, no cutoff applied. (default: %(default)s)')
    parser.add_argument('--max-length', default=0, type=int, 
                        help='Only apply filtering to contigs < max-length.\n'
                             'Longer contigs are retained, regardless of the prediction score.\n'
                             'If <1, no cutoff applied. (default: %(default)s)')
    parser.add_argument('--name-regex', default=None, type=str,
                        help='regex to remove text from the fasta sequence IDs'
                        ' (default: %(default)s).\n'
                        'For example: "_CONTIG[0-9]+$"')
    parser.add_argument('--n-proc', default=1, type=int, 
                        help='Number of fasta files to process in parallel (default: %(default)s)')
    # test args
    if test_args:
        args = parser.parse_args(test_args)
        return args
    # return
    return parser

def main(args=None):
    logging.basicConfig(format='%(asctime)s - %(message)s', level=logging.DEBUG)
    # Input
    if args is None:
        args = parse_args()
        print()
        print (args)
        print()
    # Main interface
    filter_contigs.main(args)
    
# main
if __name__ == '__main__':
    pass





import csv
import gzip
import json
import logging
import math
import mmap
import os
import statistics
import struct

from glob import glob
from timeit import default_timer as timer
from typing import Dict, List, Tuple

import numpy as np
from resmico import reader


def _replace_with_nan(data, feature_name, v):
    """Replaces all elements in arr that are equal to v with np.nan"""
    if feature_name not in data:
        return
    arr = data[feature_name]
    arr[arr == v] = np.nan


def _read_feature(file: gzip.GzipFile, data, feature_name: str, bytes: int, dtype, feature_names: List[str],
                  normalize_by: int = 1):
    if feature_name not in feature_names:
        file.seek(bytes, os.SEEK_CUR)
        return
    # `.astype()` is needed even if dtype already is np.float32, otherwise the array is read-only
    data[feature_name] = np.frombuffer(file.read(bytes), dtype=dtype).astype(np.float32)
    if normalize_by != 1:
        data[feature_name] /= normalize_by


def _to_float_and_normalize(features, result, feature_name, normalize_by):
    if feature_name in features:
        result[feature_name] = features[feature_name].astype(np.float32)
        result[feature_name] /= normalize_by


def _to_float_and_nan(features, result, feature_name, nan_value):
    if feature_name in features:
        result[feature_name] = features[feature_name].astype(np.float32)
        result[feature_name][result[feature_name] == nan_value] = np.nan


def _assign(result, orig, feature_names: List[str]):
    for feature_name in feature_names:
        if feature_name in orig:
            result[feature_name] = orig[feature_name]


def _post_process_features(features):
    result = {}
    if 'ref_base' in features:
        ref_base = features['ref_base']
        result['ref_base_A'] = np.where(ref_base == 65, 1, 0)
        result['ref_base_C'] = np.where(ref_base == 67, 1, 0)
        result['ref_base_G'] = np.where(ref_base == 71, 1, 0)
        result['ref_base_T'] = np.where(ref_base == 84, 1, 0)

    # coverage is uint16, but it needs to be cast to flaot32 because it's going to be normalized by mean/stdev later
    if 'coverage' in features:
        result['coverage'] = features['coverage'].astype(np.float32)
    _to_float_and_normalize(features, result, 'num_query_A', 10000)
    _to_float_and_normalize(features, result, 'num_query_C', 10000)
    _to_float_and_normalize(features, result, 'num_query_G', 10000)
    _to_float_and_normalize(features, result, 'num_query_T', 10000)
    _to_float_and_normalize(features, result, 'num_SNPs', 10000)
    _to_float_and_normalize(features, result, 'num_discordant', 10000)
    _to_float_and_normalize(features, result, 'num_proper_Match', 10000)
    _to_float_and_normalize(features, result, 'num_orphans_Match', 10000)
    _to_float_and_normalize(features, result, 'num_proper_SNP', 10000)

    _to_float_and_nan(features, result, 'min_insert_size_Match', 65535)
    _to_float_and_nan(features, result, 'max_insert_size_Match', 65535)
    _to_float_and_nan(features, result, 'min_mapq_Match', 255)
    _to_float_and_nan(features, result, 'max_mapq_Match', 255)
    _to_float_and_nan(features, result, 'min_al_score_Match', 127)
    _to_float_and_nan(features, result, 'max_al_score_Match', 127)

    _assign(result, features, [
        'mean_insert_size_Match',
        'stdev_insert_size_Match',
        'mean_mapq_Match',
        'stdev_mapq_Match',
        'mean_al_score_Match',
        'stdev_al_score_Match',
        'seq_window_perc_gc',
        'seq_window_entropy'])

    return result


def _read_contig_data(input_file, feature_names: List[str]):
    """
    Read a binary gzipped file containing the features for a single contig, as written by bam2feat. Features that don't
    exist are silently ignored.
    Parameters:
         - input_file: the file to read, opened at the correct offset
         - feature_names list of feature names to return (e.g. ['coverage', 'num_discordant', 'min_mapq_Match'])
    Returns:
         - a map from feature name to feature data
    """
    data = {}
    with gzip.open(input_file) as f:
        contig_size = struct.unpack('I', f.read(4))[0]
        ref_base = np.frombuffer(f.read(contig_size), dtype=np.uint8)
        # create the one-hot encoding for the reference base
        data['ref_base_A'] = np.where(ref_base == 65, 1, 0)
        data['ref_base_C'] = np.where(ref_base == 67, 1, 0)
        data['ref_base_G'] = np.where(ref_base == 71, 1, 0)
        data['ref_base_T'] = np.where(ref_base == 84, 1, 0)
        data['coverage'] = np.frombuffer(f.read(2 * contig_size), dtype=np.uint16).astype(np.float32)
        # everything is converted to float32, because frombuffer creates an immutable array, so the int values need to
        # be made mutable (in order to convert from fixed point back to float) and the float values need to be copied
        # (in order to make them writeable for normalization)
        _read_feature(f, data, 'num_query_A', 2 * contig_size, np.uint16, feature_names, 10000)
        _read_feature(f, data, 'num_query_C', 2 * contig_size, np.uint16, feature_names, 10000)
        _read_feature(f, data, 'num_query_G', 2 * contig_size, np.uint16, feature_names, 10000)
        _read_feature(f, data, 'num_query_T', 2 * contig_size, np.uint16, feature_names, 10000)
        _read_feature(f, data, 'num_SNPs', 2 * contig_size, np.uint16, feature_names, 10000)
        _read_feature(f, data, 'num_discordant', 2 * contig_size, np.uint16, feature_names, 10000)

        _read_feature(f, data, 'min_insert_size_Match', 2 * contig_size, np.uint16, feature_names)
        _read_feature(f, data, 'mean_insert_size_Match', 4 * contig_size, np.float32, feature_names)
        _read_feature(f, data, 'stdev_insert_size_Match', 4 * contig_size, np.float32, feature_names)
        _read_feature(f, data, 'max_insert_size_Match', 2 * contig_size, np.uint16, feature_names)
        _replace_with_nan(data, 'min_insert_size_Match', 65535)
        _replace_with_nan(data, 'max_insert_size_Match', 65535)

        _read_feature(f, data, 'min_mapq_Match', contig_size, np.uint8, feature_names)
        _read_feature(f, data, 'mean_mapq_Match', 4 * contig_size, np.float32, feature_names)
        _read_feature(f, data, 'stdev_mapq_Match', 4 * contig_size, np.float32, feature_names)
        _read_feature(f, data, 'max_mapq_Match', contig_size, np.uint8, feature_names)
        _replace_with_nan(data, 'min_mapq_Match', 255)
        _replace_with_nan(data, 'max_mapq_Match', 255)

        _read_feature(f, data, 'min_al_score_Match', contig_size, np.int8, feature_names)
        _read_feature(f, data, 'mean_al_score_Match', 4 * contig_size, np.float32, feature_names)
        _read_feature(f, data, 'stdev_al_score_Match', 4 * contig_size, np.float32, feature_names)
        _read_feature(f, data, 'max_al_score_Match', contig_size, np.int8, feature_names)
        _replace_with_nan(data, 'min_al_score_Match', 127)
        _replace_with_nan(data, 'max_al_score_Match', 127)

        _read_feature(f, data, 'num_proper_Match', 2 * contig_size, np.uint16, feature_names, 10000)
        _read_feature(f, data, 'num_orphans_Match', 2 * contig_size, np.uint16, feature_names, 10000)
        _read_feature(f, data, 'num_proper_SNP', 2 * contig_size, np.uint16, feature_names, 10000)
        _read_feature(f, data, 'seq_window_perc_gc', 4 * contig_size, np.float32, feature_names)
        _read_feature(f, data, 'seq_window_entropy', 4 * contig_size, np.float32, feature_names)
    return data


class ContigInfo:
    """
    Contains metadata about a single contig.
    """

    def __init__(self, name: str, file_name: str, length: int, offset: int, size_bytes: int, misassembly_count: int,
                 breakpoints: List[Tuple[int, int]], avg_coverage: float):
        self.name: str = name
        self.file: str = file_name
        self.length: int = length
        self.offset: int = offset
        self.size_bytes: int = size_bytes
        self.misassembly: int = misassembly_count
        self.features: Dict[str:np.array] = {}
        self.breakpoints = breakpoints
        self.avg_coverage = avg_coverage


class ContigReader:
    """
    Reads contig data from binary files written by ResMiCo-SM.
    """

    def __init__(self, input_dirs: str, feature_names: List[str], process_count: int,
                 no_cython: bool = False, stats_file: str = '', min_len: int = 0, min_avg_coverage: int = 0,
                 feature_file_match: str = ''):
        """
        Arguments:
            - input_dirs: location on disk where the feature data is stored
            - feature_names: feature names to use in training
            - process_count: number of processes to use for loading data in parallel
            - no_cython: whether to read data from disk using pure Python or using Cython bindings
            - stats_file: if present, specifies a stats file to read the statistics for each feature from
            - min_len: exclude all contigs shorter than min_len
            - input_dir_match string that the directories for the feature files must match; useful for filtering
              contigs with certain properties, such as sequencing depth, abundance, etc.
        """
        # means and stdevs are a map from feature name to the mean and standard deviation precomputed for that
        # feature across *all* contigs, stored as a tuple
        self.means: Dict[str, float] = {}
        self.stdevs: Dict[str, float] = {}
        # a list of ContigInfo objects with metadata about all contigs found in #input_dir
        self.contigs: List[ContigInfo] = []

        self.feature_names = feature_names
        self.process_count = process_count
        self.no_cython = no_cython
        
        # just temp attributes for measuring performance
        self.normalize_time = 0
        self.read_time = 0
        self.min_len = min_len
        self.min_avg_coverage = min_avg_coverage
        self.feature_file_match = feature_file_match

        self.feature_mask: List[int] = [1 if feature in feature_names else 0 for feature in reader.feature_names]

        # getting feature file paths
        file_list = []
        if os.path.isfile(input_dirs):
            msg = '  Assuming that the feature files are provided in a file as a list'
            logging.info(msg)
            base_dir = os.path.split(input_dirs)[0]
            with open(input_dirs) as inF:
                for line in inF:
                    line = line.rstrip().split(' ')[0]
                    if os.path.isfile(line):
                        file_list.append(line)
                    else:
                        # appending on the path of the file list
                        line = os.path.join(base_dir, line)
                        if os.path.isfile(line):
                            file_list.append(line)
        else:
            logging.info('Looking for stats/toc files...')
            for input_dir in input_dirs.split(','):
                fl = list(glob(os.path.join(input_dir, '**/stats'), recursive=True))
                fx = os.path.join(input_dir, 'stats')
                if os.path.isfile(fx) and not fx in fl:
                    fl.append(fx)
                if feature_file_match:
                    count = len(fl)
                    fl = [f for f in fl if feature_file_match in f]
                    logging.info(
                        f'Filtered for directories matching: {feature_file_match}. {len(file_list)} out of {count} kept')
                file_list.extend(fl)        
                logging.info(f'Processing {len(file_list)} stats/toc files found in {input_dir} ...')        
        if not file_list:
            logging.info('Nothing to do.')
            exit(0)
        elif ',' in input_dirs:
            logging.info(f'A total of {len(file_list)} stats/toc files found')

        # stats.json file
        if stats_file == '':
            logging.info('Computing global means and standard deviations...')
            self._compute_mean_stdev(file_list)
            if ',' not in input_dirs:
                if os.path.isfile(input_dirs):
                    out_file = os.path.join(os.path.split(input_dirs)[0], 'stats.json')
                else:
                    out_file = os.path.join(input_dirs, 'stats.json')
                json.dump({'means': self.means, 'stdevs': self.stdevs}, open(out_file, 'w'), indent=2)
                logging.info(f'Means and stdevs saved to: {out_file}')
            else:
                logging.info('Composed input dir: not writing stats.json')
        else:
            logging.info(f'Loading feature means and standard deviations from {stats_file}')
            means_stdevs = json.load(open(stats_file))
            self.means, self.stdevs = means_stdevs['means'], means_stdevs['stdevs']

        self._load_contigs_metadata(file_list)    
        
    def __len__(self):
        return len(self.contigs)

    def read_contigs(self, contig_infos: List[ContigInfo], return_raw=False):
        """
        Reads the features for the given contig_infos from file and returns the result in a list of len(contig_infos)
        dictionaries of {'feature_name', feature_data}
        """
        start = timer()
        self.normalize_time = 0
        self.read_time = 0
        result = []
        if self.no_cython:
            for contig_data in map(self._read_and_normalize, contig_infos):
                result.append(contig_data)
        else:
            file_names: List[bytes] = []
            lengths: List[int] = []
            offsets: List[int] = []
            sizes: List[int] = []
            for c in contig_infos:
                file_names.append(c.file.encode('utf-8'))
                lengths.append(c.length)
                offsets.append(c.offset)
                sizes.append(c.size_bytes)

            features_raw = reader.read_contigs_py(file_names, lengths, offsets, sizes, self.feature_mask,
                                                  self.process_count)
            # traverse features for each contig, convert to proper data type and normalize by mean/stdev if needed
            for f in features_raw:
                features = _post_process_features(f)
                if not return_raw:
                    self._normalize(features)
                result.append(features)
        logging.debug(f'Contigs read in {(timer() - start):5.2f}s; read: {self.read_time:5.2f}s '
                      f'normalize: {self.normalize_time:5.2f}s')
        return result

    def _compute_mean_stdev(self, file_list):
        mean_count = 0
        stddev_count = 0
        all_count = 0
        metrics = ['insert_size', 'mapq', 'al_score']  # insert size, mapping quality, alignment quality
        metric_types = ['min', 'mean', 'stdev', 'max']

        # test and see if the new metrics (coverage, gc, entropy) are already in the stats (either because
        # the new bam2feat version was used or because they were added via add_stats.py
        # TODO: remove when all datasets have the new metrics
        has_new_metrics = False
        with open(file_list[0]) as f:
            stats = json.load(f)
            if 'all_count' in stats and 'seq_window_perc_gc' in stats and 'seq_window_entropy' in stats and 'coverage' in stats:
                has_new_metrics = True

        has_new_metrics = False  # TODO: remove
        new_metrics = ['coverage', 'seq_window_entropy', 'seq_window_perc_gc'] if has_new_metrics else []

        for feature_name in reader.float_feature_names + new_metrics:
            self.means[feature_name] = 0
            self.stdevs[feature_name] = 0

        for fname in file_list:
            with open(fname) as f:
                stats = json.load(f)
                if 'all_count' in stats:
                    all_count += stats['all_count']
                mean_count += stats['mean_cnt']
                stddev_count += stats['stdev_cnt']
                for metric in metrics:
                    for mtype in metric_types:
                        feature_name = f'{mtype}_{metric}_Match'
                        self.means[feature_name] += stats[metric]['sum'][mtype]
                        self.stdevs[feature_name] += stats[metric]['sum2'][mtype]
                for metric in new_metrics:
                    self.means[metric] += stats[metric]['sum']
                    self.stdevs[metric] += stats[metric]['sum2']

        for metric in metrics:
            for mtype in metric_types:
                feature_name = f'{mtype}_{metric}_Match'
                # the count of non-NaN position is different (lower) for the stdev_* features (because computing
                # the standard deviation requires at least coverage 2, while for max/mean/min coverage=1 is sufficient)
                cnt = stddev_count if mtype == 'stdev' else mean_count
                var = cnt * self.stdevs[feature_name] - self.means[feature_name] ** 2
                if -0.1 < var < 0:  # fix small rounding errors that may lead to negative values
                    var = 0
                self.stdevs[feature_name] = math.sqrt(var) / cnt
                self.means[feature_name] /= cnt
        for feature_name in new_metrics:
            cnt = all_count
            var = cnt * self.stdevs[feature_name] - self.means[feature_name] ** 2
            if -0.1 < var < 0:  # fix small rounding errors that may lead to negative values
                var = 0
            self.stdevs[feature_name] = math.sqrt(var) / cnt
            self.means[feature_name] /= cnt
        # print the computed values
        header = ''
        values = ''
        for metric in metrics:
            for mtype in metric_types:
                feature_name = f'{mtype}_{metric}_Match'
                header += f'{feature_name}\t\t\t'
                values += f'{self.means[feature_name]:.2f}/{self.stdevs[feature_name]}\t\t\t'
        for feature_name in new_metrics:
            header += f'{feature_name}\t\t\t'
            values += f'{self.means[feature_name]:.2f}/{self.stdevs[feature_name]}\t\t\t'
        separator = '_' * 300
        logging.info(
            'Computed global means and stdevs:\n' + separator + '\n' + header + '\n' + values + '\n' + separator)

    def _load_contigs_metadata(self, file_list):
        contig_count = 0
        contig_count_misassembled = 0
        total_len = 0
        breakpoint_hist = np.zeros(50, np.int32)
        breakpoint_relpos_hist = np.zeros(20, np.int32)
        excluded_count = 0
        contig_lengths = []
        for fname in file_list:
            toc_file = fname[:-len('stats')] + 'toc'
            contig_fname = fname[:-len('stats')] + 'features_binary'
            offset = 0
            with open(toc_file) as f:
#                 logging.info(f'FILE: {toc_file}')
                rd = csv.reader(f, delimiter='\t', quoting=csv.QUOTE_NONE)
                try: #file is not empty
                    next(rd)  # skip CSV header: Assembler, Contig_name, MissassembleCount, ContigLen
                except:
                    continue
                for row in rd:
                    # the fields in row are: name, length (bases), misassembly_count, size_bytes, breakpoints, coverage
                    contig_name = row[0]
                    contig_len = int(row[1])
                    size_bytes = int(row[3])
                    breakpoints = []
                    if len(row) >= 5:  # breakpoints is present; TODO: remove this once all datasets have it
                        if row[4] != '-':
                            all_breakpoints = row[4].split(',')
                            for break_point in all_breakpoints:
                                start_stop = break_point.split('-')
                                breakpoints.append((int(start_stop[0]), int(start_stop[1])))
                                mid = (int(start_stop[0]) + int(start_stop[1])) // 2
                                # since contigs can be reversed, chose the breakpoint closer to the edge
                                # (or simply duplicate the breakpoint in case of relative position histogram)
                                breakpoint_relpos_hist[int(mid * 20 / contig_len)] += 1
                                breakpoint_relpos_hist[int((contig_len - mid) *20 / contig_len)] += 1
                                if contig_len - mid < mid:
                                    mid = contig_len - mid
                                breakpoint_hist[min(49, mid // 200)] += 1
                    avg_coverage = float(row[5]) if len(row) >= 6 else 100

                    contig_info = ContigInfo(contig_name, contig_fname, contig_len, offset, size_bytes, int(row[2]),
                                             breakpoints, avg_coverage)
                    if contig_info.length >= self.min_len and contig_info.avg_coverage >= self.min_avg_coverage:
                        contig_lengths.append(contig_info.length)
                        total_len += contig_info.length
                        self.contigs.append(contig_info)
                    else:
                        excluded_count += 1
                    offset += size_bytes
                    if contig_info.misassembly:
                        contig_count_misassembled += 1
                    contig_count += 1
        logging.info(
            f'Found {contig_count} contigs, {contig_count_misassembled} misassembled, {excluded_count} excluded, '
            f'{total_len} total length, {statistics.median(contig_lengths)} median length, '
            f'memory needed (assuming fraq-neg=1) {total_len * reader.bytes_per_base / 1e9:6.2f}GB')
        logging.info(f'Breakpoint location histogram: {",".join([str(x) for x in breakpoint_hist])}')
        logging.info(f'Breakpoint relative position histogram: {",".join([str(x) for x in breakpoint_relpos_hist])}')

    def read_file(self, fname):
        toc_file = fname[:-len('stats')] + 'toc'
        contig_fname = fname[:-len('stats')] + 'features_binary'
        offset = 0
        result = []
        with open(toc_file) as f, open(contig_fname) as binary_file:
            binary_file = open(contig_fname, 'rb')
            # memory-map the file, size 0 means whole file
            mm = mmap.mmap(binary_file.fileno(), 0, access=mmap.ACCESS_READ)

            rd = csv.reader(f, delimiter='\t', quoting=csv.QUOTE_NONE)
            try:
                next(rd)  # skip CSV header: Assembler, Contig_name, MissassembleCount, ContigLen
            except:
                return
            for row in rd:
                size_bytes = int(row[3])
                # the gzip reader reads ahead and messes up the current position, so we need to re-seek
                mm.seek(offset)
                features = _read_contig_data(mm, self.feature_names)
                self._normalize(features)
                result.append(features)
                offset += size_bytes
        return result

    def _read_and_normalize(self, contig_info: ContigInfo):
        """
        Reads and normalizes the float features present in features using the precomputed means and standard deviations
        in #mean_stdev
        Parameters:
            contig_info: the metadata of the contig to be loaded
        """
        if contig_info.features:  # data is cached in memory, simply return cached value
            return contig_info.features
        start = timer()

        # features is a map from feature name (e.g. 'coverage') to a numpy array containing the feature

        input_file = open(contig_info.file, mode='rb')
        input_file.seek(contig_info.offset)
        features = _read_contig_data(input_file, self.feature_names)

        self.read_time += (timer() - start)
        self._normalize(features)
        return features

    def _normalize(self, features):
        start = timer()
        has_new_metric = False
        if 'coverage' in self.means and 'coverage' in self.stdevs:
            has_new_metric = True
        new_metric = ['coverage'] if has_new_metric else []
        for feature_name in reader.float_feature_names + new_metric:
            if feature_name not in features:
                continue
            if feature_name not in self.means or feature_name not in self.stdevs:
                logging.warning(
                    f'Could not find mean/standard deviation for feature: {feature_name}. Skipping normalization')
                continue
            features[feature_name] -= self.means[feature_name]

            if features[feature_name].dtype != np.float32:  # we can do the division in place
                features[feature_name] = features[feature_name].astype(np.float32)
            if self.stdevs[feature_name] > 1e-3:
                features[feature_name] /= self.stdevs[feature_name]
            else:
                continue
                logging.warning(f'Stdev for {feature_name} is too low ({self.stdevs[feature_name]}). Not normalizing')
            #             replace NANs with 0 (the new mean)
            nan_pos = np.isnan(features[feature_name])
            features[feature_name][nan_pos] = 0
        self.normalize_time += (timer() - start)


if __name__ == '__main__':
    pass




import argparse
import logging
import sys
from resmico import train_binary_data
from resmico.commands import arguments


def parse_args(curr_args=None, subparsers=None):
    desc = "Train a new model using resmico"
    if subparsers:
        parser = subparsers.add_parser('train', description=desc,
                                       formatter_class=argparse.RawTextHelpFormatter)
    else:
        parser = argparse.ArgumentParser(description=desc,
                                         formatter_class=argparse.RawTextHelpFormatter)
    parser_g1 = parser.add_argument_group('Training-specific arguments')
    parser_g1.add_argument('--early-stop', action='store_true', default=False,
                           help='Use early stopping (default: %(default)s)')
    parser_g1.add_argument('--net-type', default='cnn_resnet_avg', type=str,
                           help='ResMiCo NN type (default: %(default)s)')
    parser_g1.add_argument('--num-blocks', default=4, type=int,
                           help='Number of residual blocks (3 or 4, 5, 6) (default: %(default)s)')
    parser_g1.add_argument('--filters', default=16, type=int,
                           help='N of filters for first conv layer. Then x2 (default: %(default)s)')
    parser_g1.add_argument('--ker-size', default=5, type=int,
                           help='CNN kernel size (default: %(default)s)')
    parser_g1.add_argument('--n-hid', default=50, type=int,
                           help='N of units in fully connected layers (default: %(default)s)')
    parser_g1.add_argument('--n-conv', default=5, type=int,
                           help='N of conv layers (default: %(default)s)')
    parser_g1.add_argument('--n-fc', default=1, type=int,
                           help='N of fully connected layers (default: %(default)s)')
    parser_g1.add_argument('--n-epochs', default=50, type=int,
                           help='N of training epochs (default: %(default)s)')
    parser_g1.add_argument('--batch-size', default=16, type=int,
                           help='Batch size (default: %(default)s)')
    parser_g1.add_argument('--dropout', default=0, type=float,
                           help='Rate of dropout (default: %(default)s)')
    parser_g1.add_argument('--lr-init', default=0.0001, type=float,
                           help='Initial learning rate')
    parser_g1.add_argument('--fraq-neg', default=1., type=float,
                           help='Portion of samples to keep in overrepresented class (default: %(default)s)')
    parser_g1.add_argument('--cache', dest='cache', action='store_true',
                           help='Cache the train & validation data in memory for quicker access (default: %(default)s)')
    parser_g1.add_argument('--cache-validation', dest='cache_validation', action='store_true',
                           help='Cache the validation data in memory for quicker access (default: %(default)s)')
    parser_g1.add_argument('--cache-train', dest='cache_train', action='store_true',
                           help='Cache the train data in memory for quicker access (default: %(default)s)')
    parser_g1.add_argument('--log-progress', default=False,
                           help='Show a progressbar for training/evaluation progress (default: %(default)s)',
                           dest='log_progress', action='store_true')
    parser_g1.add_argument('--num-translations', default=1, type=int,
                           help='How many variations to select for each sample (default: %(default)s)\n'
                           'For positive samples, the variation will be around the breaking point)')
    parser_g1.add_argument('--max-translation-bases', default=0, type=int,
                           help='Maximum number of bases to translate around the breaking point\n'
                           '(i.e. misassembled contigs) that are longer than max-len (default: %(default)s)')
    parser_g1.add_argument('--weight-factor', default=0, type=int,
                           help='Factor by which contigs are weighted based on their length (default: %(default)s).\n'
                           'w = min(1, contig_len/weight-factor). 0==no weighting')
    parser_g1.add_argument('--model-checkpoint', default='', type=str,
                           help='Path to the trained model to continue the training')

    arguments.add_common_args(parser)

    # running test args
    if curr_args:
        args = parser.parse_args(curr_args)
        return args

    return parser


def main(args=None):
    """
    Main interface
    """
    if args is None:
        args = parse_args(sys.argv[1:])
    logging.basicConfig(format='%(asctime)s - %(message)s', level=logging._nameToLevel[args.log_level.upper()])
    train_binary_data.main(args)


# main
if __name__ == '__main__':
    main()



import os
import re
import bz2
import gzip
import logging
import multiprocessing as mp
from functools import partial
from resmico import utils

def _open(infile, mode='rb'):
    """
    Openning of input, regardless of compression
    """
    if infile.endswith('.bz2'):
        return bz2.open(infile, mode)
    elif infile.endswith('.gz'):
        return gzip.open(infile, mode)
    else:
        return open(infile)

def _decode(x):
    """
    Decoding input, if needed
    """
    try:
        x = x.decode('utf-8')
    except AttributeError:
        pass
    return x

def parse_predictions(infile, delim=',', name_regex=None):
    """
    Parsing contig prediction scores
    """
    logging.info('Parsing predictions table...')
    if name_regex is not None:
        regex = re.compile(name_regex)
    else:
        regex = None
    header = dict()
    predictions = dict()
    with _open(infile) as inF:
        for i,line in enumerate(inF):
            line = _decode(line).rstrip().split(delim)
            # header
            if i == 0:
                header = {x:ii for ii,x in enumerate(line)}
                for col in ['cont_name', 'score', 'length']:
                    try:
                        _ = header[col]
                    except KeyError:
                        raise KeyError(f'Cannot find required column: "{col}"')
                continue
            # body
            contig_id = os.path.split(line[header['cont_name']])[1]
            if regex is not None:
                contig_id = regex.sub('', contig_id)
            score = float(line[header['score']])
            contig_len = float(line[header['length']])
            predictions[contig_id] = [score, contig_len]
    # status
    n_obs = len(predictions.keys())
    logging.info(f'  No. of predictions: {n_obs}')
    scores = [x[0] for x in predictions.values()]
    min_score = round(min(scores),2)
    mean_score = round(sum(scores) / n_obs,2)
    max_score = round(max(scores),2)
    logging.info(f'  Min score: {min_score}; Mean score: {mean_score}; Max score: {max_score}')
    return predictions

def score_cutoff_check(score_cutoff):
    if score_cutoff < 0:
        logging.warning('WARNING: --score-cutoff < 0')
    elif score_cutoff > 1:
        logging.warning('WARNING: --score-cutoff > 1')

def check_files_exist(in_files):
    for F in in_files:
        if not os.path.exists(F):
            raise IOError(f'Cannot find file: {F}')
        
def parse_fasta_list(fasta_list):
    """
    If a list of fasta files (1 per line), parsing list & returning.
    Otherwise, assuming input is the list of fasta file paths.
    """
    if len(fasta_list) > 1:
        return fasta_list
    fasta_files = []
    with _open(fasta_list[0]) as inF:
        for line in inF:
            line = _decode(line.rstrip())
            if line == '':
                # skipping blank lines
                continue
            if line.startswith('>'):
                # assumign fasta
                fasta_files = fasta_list
                break
            else:
                # assuming file
                if not os.path.exists(line):
                    # assuming that genomes in the same directory as the list file
                    d = os.path.split(fasta_list[0])[0]
                    line = os.path.join(d, line)
                    if not os.path.exists(line):
                        raise IOError(f'Cannot find file: {line}')
                # appending paths to list
                fasta_files.append(line)    
    return fasta_files

def _filter_fasta(fasta_file, predictions, outdir, pred_score_cutoff, add_score=False,
                  error_on_missing=False, min_length=0, max_length=0, outfile=None):
    """
    Filtering fasta based on contig prediction scores
    """    
    logging.info(f'Filtering fasta: {fasta_file}')
    if outfile is None:
        outfile = os.path.join(outdir, os.path.split(fasta_file)[1])
    else:
        outfile = os.path.join(outdir, os.path.split(outfile)[1])
    if outfile == fasta_file:
        msg = 'Input path cannot equal output path: {} <=> {}'
        raise ValueError(msg.format(fasta_file, outfile))
    gzip_out = outfile.endswith('.gz')        
    contig_id = None
    pred_score = None
    contig_len = None
    stats = {'kept':0, 'filtered':0, 'total':0, 'match':0}
    if gzip_out:
        outF = gzip.open(outfile, 'wb')
    else:
        outF = open(outfile, 'w')
    with _open(fasta_file) as inF:
        for i,line in enumerate(inF):
            line = _decode(line).rstrip()
            if line == '':
                continue
            # seq header
            if line.startswith('>'):
                stats['total'] += 1
                contig_id = line.lstrip('>')
                # getting score for contig
                try:
                    pred_score = float(predictions[contig_id][0])
                    contig_len = float(predictions[contig_id][1])
                    stats['match'] += 1
                except KeyError:
                    msg = 'Cannot find contig "{}" in the list of predictions'
                    if error_on_missing is True:
                        raise ValueError(msg.format(contig_id))
                    pred_score = 0
                # status
                if min_length > 0 and contig_len < max_length:
                    # retaining short contigs, regardless of score
                    stats['kept'] += 1
                    pred_score = 0                    
                elif max_length > 0 and contig_len > max_length:
                    # retaining long contigs, regardless of score
                    stats['kept'] += 1
                    pred_score = 0
                elif pred_score is not None and pred_score < pred_score_cutoff:
                    # retaining contig due to low score
                    stats['kept'] += 1
                else:
                    # score hits cutoff; filtering
                    stats['filtered'] += 1
                    continue
                # add score to sequence header?
                if add_score is True:
                    line = '>{} score={}'.format(contig_id, round(pred_score,6))
            # applying prediction score cutoff
            if pred_score is not None and pred_score < pred_score_cutoff:
                line = line + '\n'
                if gzip_out:
                    outF.write(line.encode('utf-8'))
                else:
                    outF.write(line)    
    # status
    outF.close()
    logging.info('  No. of input contigs: {}'.format(stats['total']))
    logging.info('  No. of contigs with a prediction: {}'.format(stats['match']))
    logging.info('  No. of filtered contigs: {}'.format(stats['filtered']))
    logging.info('  No. of output contigs: {}'.format(stats['kept']))
    logging.info(f'  File written: {outfile}')

def set_logger(level):
    logger = logging.getLogger()
    logger.setLevel(level)
    for handler in logger.handlers:
        handler.setLevel(level)
        
def filter_fasta(fasta_files, predictions, outdir, pred_score_cutoff,
                 add_score=False, error_on_missing=False,
                 min_length=0, max_length=0, n_proc=1, outfile=None):
    """
    Filter >=1 fasta based on prediction scores
    """
    if not os.path.isdir(outdir):
        os.makedirs(outdir)
    func = partial(_filter_fasta,
                   predictions = predictions,
                   outdir = outdir,
                   pred_score_cutoff = pred_score_cutoff,
                   add_score = add_score,
                   error_on_missing = error_on_missing,
                   min_length = min_length,
                   max_length = max_length,
                   outfile = outfile)
    if n_proc < 2:
        res = map(func, fasta_files)
    else:
        logging.info('Filtering fasta files in parallel...')
        set_logger(logging.WARNING)
        pool = mp.Pool(n_proc)
        res = pool.map(func, fasta_files)
        set_logger(logging.INFO)
    return [x for x in res]
    
def main(args):
    # fasta file(s)
    args.fasta = parse_fasta_list(args.fasta)
    logging.info('No. of fasta files: {}'.format(len(args.fasta)))
    # checks
    score_cutoff_check(args.score_cutoff)
    # predictions
    predictions = parse_predictions(args.prediction_table,
                                    args.score_delim,
                                    args.name_regex)
    # fasta filter
    filter_fasta(args.fasta, predictions, args.outdir, args.score_cutoff,
                 args.add_score, args.error_on_missing, args.min_length,
                 args.max_length, args.n_proc, args.outfile)

if __name__ == '__main__':
    pass



import _pickle as pickle
import os
import sys
import csv
import gzip
from pathlib import Path
import logging
from toolz import itertoolz
from functools import partial
from collections import defaultdict
import multiprocessing as mp
import tables
import pathos
import time

import numpy as np
import pandas as pd
from sklearn.metrics import average_precision_score, recall_score
from sklearn.model_selection import train_test_split
from sklearn.isotonic import IsotonicRegression
from tensorflow.keras import backend as K
import tensorflow as tf
from tensorflow.keras.callbacks import Callback
from tensorflow.keras.layers import Conv1D, ReLU, BatchNormalization, Add, Cropping1D, AveragePooling1D
from tensorflow.keras.layers import Masking, LayerNormalization, MultiHeadAttention, Dropout

from resmico import contig_reader
from resmico import models_fl as Models

def class_recall_0(y_true, y_pred):
    label = 0
    class_id_preds = K.cast(K.greater(y_pred, 0.5), 'int32')
    y_true = K.cast(y_true, 'int32')
    accuracy_mask = K.cast(K.equal(y_true, label), 'int32')
    class_acc_tensor = K.cast(K.equal(y_true, class_id_preds), 'int32') * accuracy_mask
    class_acc = K.sum(class_acc_tensor) / K.maximum(K.sum(accuracy_mask), 1)
    return class_acc


def class_recall_1(y_true, y_pred):
    label = 1
    class_id_preds = K.cast(K.greater(y_pred, 0.5), 'int32')
    y_true = K.cast(y_true, 'int32')
    accuracy_mask = K.cast(K.equal(y_true, label), 'int32')
    class_acc_tensor = K.cast(K.equal(y_true, class_id_preds), 'int32') * accuracy_mask
    class_acc = K.sum(class_acc_tensor) / K.maximum(K.sum(accuracy_mask), 1)
    return class_acc


'''
Constructs the essential building unit of Resmico: the residual block. 
'''
def relu_bn(inputs):
    relu = ReLU()(inputs)
    bn = BatchNormalization()(relu)
    return bn

def bn_relu(inputs):
    bn = BatchNormalization()(inputs)
    relu = ReLU()(bn)
    return relu

# ### 1st version with deterministic predictions
def old_residual_block(x, downsample: bool, filters, kernel_size):
    y = Conv1D(kernel_size=kernel_size,
               strides=(1 if not downsample else 2),
               filters=filters,
               padding='valid' if downsample else 'same')(x)
    y = relu_bn(y)
    y = Conv1D(kernel_size=kernel_size,
               strides=1,
               filters=filters,
               padding='valid')(y)

    if downsample:
        x = Conv1D(kernel_size=1,
                   strides=2,
                   filters=filters,
                   padding='valid')(x)
        # the additional cropping is needed in order to match the size of the y=Conv1D() output, since here we
        # user kernel_size=1
        x = Cropping1D((0,kernel_size//2))(x)
    x = Cropping1D((0, kernel_size-1))(x)
    out = Add()([x, y])
    out = relu_bn(out)
    return out

'''
Residual block with all valid pading
'''
def residual_block(x, downsample: bool, filters, kernel_size):
    y = Conv1D(kernel_size=kernel_size,
               strides=(1 if not downsample else 2),
               filters=filters,
               padding='valid')(x)
    y = relu_bn(y)
    y = Conv1D(kernel_size=kernel_size,
               strides=1,
               filters=filters,
               padding='valid')(y)

    if downsample:
        x = Conv1D(kernel_size=1,
                   strides=2,
                   filters=filters,
                   padding='valid')(x)
        # the cropping is needed to make up for the size difference caused by the convolutions on y
        x = Cropping1D((0,kernel_size-1 + kernel_size//2))(x)

    else:
        x = Cropping1D((0, 2*kernel_size-2))(x)

    out = Add()([x, y])
    out = relu_bn(out)
    return out


def dilated_residual_block(x, dilate: bool, filters, kernel_size):
    ### if dilate==True: apply dilation rate 3

    if filters>64:
        #bottleneck layer
        y = Conv1D(kernel_size=1,
               filters=64,
               padding='valid')(x)
        y = bn_relu(y)
        y = Conv1D(kernel_size=kernel_size,
               dilation_rate=(1 if not dilate else 3),
               filters=filters,
               padding='valid')(y)
        y = bn_relu(y)
        y = Conv1D(kernel_size=1,
               filters=filters,
               padding='valid')(y)
        y = BatchNormalization()(y)
        
        if dilate: 
            x = Conv1D(kernel_size=1,
                       filters=filters,
                       padding='valid')(x)
            x = Cropping1D((0,((kernel_size-1)*3)))(x)
        else:
            x = Cropping1D((0, (kernel_size-1)))(x)
    
    else:
        #residual block with two weighted layers
        y = Conv1D(kernel_size=kernel_size,
                   dilation_rate=(1 if not dilate else 3),
                   filters=filters,
                   padding='valid')(x)
        y = bn_relu(y)
        y = Conv1D(kernel_size=kernel_size,
                   dilation_rate=1,
                   filters=filters,
                   padding='valid')(y)
        y = BatchNormalization()(y)
        if dilate: 
            #also number of filters changed, so conv layer needed
            x = Conv1D(kernel_size=1,
                       filters=filters,
                       padding='valid')(x)
            x = Cropping1D((0,(kernel_size-1)*3 + (kernel_size-1)))(x)
        else:
            x = Cropping1D((0, 2*kernel_size-2))(x)
    out = Add()([x, y])
    out = ReLU()(out)
    return out


def transformer_encoder(inputs, head_size, num_heads, dropout=0):
    # Normalization and Attention
    x = LayerNormalization()(inputs)
    x = MultiHeadAttention(
        key_dim=head_size, num_heads=num_heads, dropout=dropout
    )(x, x)
    x = Dropout(dropout)(x)
    res = x + inputs

    # Feed Forward Part
    x = LayerNormalization()(res)
    x = Conv1D(filters=inputs.shape[-1], kernel_size=1, activation="relu")(x)
    x = Dropout(dropout)(x)
#     x = Conv1D(filters=inputs.shape[-1], kernel_size=1)(x)
    return x + res


def n_moves_window(cont_len, window, step):
    if cont_len < window:
        return 0
    else:
        return np.ceil((cont_len - window) / step)

# look at predictions
def add_stats(df, column_name='chunk_scores'):
    chunk_scores = np.array(df[column_name])
    df['min'] = [np.min(list_scores) for list_scores in chunk_scores]
    df['mean'] = [np.mean(list_scores) for list_scores in chunk_scores]
    df['std'] = [np.std(list_scores) for list_scores in chunk_scores]
    df['max'] = [np.max(list_scores) for list_scores in chunk_scores]

    percent_names = ['p10', 'p20', 'p30', 'p40', 'p50', 'p60', 'p70', 'p80', 'p90']
    pers = np.arange(10, 100, 10)
    result_shape = (df.shape[0], len(pers))
    compute_pers = [np.percentile(list_scores, pers) for list_scores in chunk_scores]
    df[percent_names] = np.array(compute_pers).reshape(result_shape)

    return df


def aggregate_chunks(batches_list, all_lens, all_labels, all_names, all_preds, window, step):
    if all_labels != []:
        dic_predictions = {'cont_name': [], 'length': [], 'label': [], 'chunk_scores': []}
    else:
        dic_predictions = {'cont_index': [], 'length': [], 'chunk_scores': []}

    start_pos = 0

    for cont_inds in batches_list:
        for cont_ind in cont_inds:
            cont_len = all_lens[cont_ind]
            dic_predictions['length'].append(cont_len)

            end_pos = start_pos + int(1 + n_moves_window(all_lens[cont_ind], window, step))
            cont_preds = all_preds[start_pos:end_pos].reshape(-1)

            # contigs of length of length 5k-6k treated the same as 5k
            if (cont_len >= 5000) & (cont_len <= 6000):
                cont_preds = cont_preds[0]
                # the second window is not informative and can be harmful, because has the same weight as the first

            dic_predictions['chunk_scores'].append(cont_preds)
            start_pos = end_pos
            if all_labels != []:
                dic_predictions['cont_name'].append(all_names[cont_ind])
                dic_predictions['label'].append(all_labels[cont_ind])

            else:
                dic_predictions['cont_index'].append(cont_ind)

    return dic_predictions


def update_progress(current: int, total: int, prefix: str, tail: str):
    """
    Displays or updates a console progress bar.
    Accepts a float between 0 and 1. Any int will be converted to a float.
    A value under 0 represents a 'halt'. A value at 1 or bigger represents 100%.
    """
    barLength = 100
    status = tail
    progress = current / total
    if progress < 0:
        progress = 0
        status = "Halt...\r\n"
    if progress >= 1:
        progress = 1
        status = "Done...\r\n"
    block = int(round(barLength * progress))
    current_str = str(current).rjust(len(str(total)))
    text = f'\r{prefix}[{"#" * block + "-" * (barLength - block)}] {current_str}/{total} {status}'
    sys.stdout.write(text)
    sys.stdout.flush()
    
def sma(arr, window_size=2):
    if len(arr)<=window_size:
        return [np.mean(arr)]
    i = 0
    # Initialize an empty list to store moving averages
    moving_averages = []
    while i < len(arr) - window_size + 1:
        window_average = np.sum(arr[
          i:i+window_size]) / window_size
        moving_averages.append(window_average)
        i += 1
    return moving_averages

def prob_score_aggr(arr):
    arr = np.array(arr)
    arr = 1-arr
    return 1 - np.prod(arr)

def calib_prob_score_aggr(scores, model):  
    return prob_score_aggr(model.predict(scores))

def test_calibprob_vc_max(model: tf.keras.Model, num_gpus: int, args):
    start = time.time()

    is_fixed_length = model.layers[0].input_shape[0][1] is not None  
    convoluted_size = Models.construct_convolution_lambda(model)
#     else:  # when not padding, the convoluted size is unused
#         convoluted_size = lambda len, pad: 0

    logging.info('Loading contig data...')
    reader = contig_reader.ContigReader(args.feature_files_path, args.features, args.n_procs,
                                        args.no_cython, args.stats_file, args.min_contig_len,
                                        min_avg_coverage=args.min_avg_coverage)

    if args.val_ind_f:
        eval_idx = list(pd.read_csv(args.val_ind_f)['val_ind'])
        logging.info(f'Using {len(eval_idx)} indices in {args.val_ind_f} for prediction')
    else:
        logging.info('Should be done on the validation set')

    #split into validation and calibration
    calib_idx, eval_idx = train_test_split(eval_idx, test_size=0.5, random_state=args.seed)
    #for calibration part filter < 20k length
    short_calib = []
    for ind in calib_idx:
        if reader.contigs[ind].length < 20000:
             short_calib.append(ind)
    calib_idx = short_calib

    
    #create predictions
    calib_data = Models.BinaryDatasetEval(reader, calib_idx, args.features, args.max_len, max(250, args.max_len), 
                                        int(args.gpu_eval_mem_gb * 1e9 * 0.8), cache_results=False,
                                        show_progress=True, convoluted_size=convoluted_size,
                                        pad_to_max_len=is_fixed_length, batch_size=args.batch_size)
    calib_data_y = np.array([0 if reader.contigs[idx].misassembly == 0 else 1 for idx in calib_data.indices])
    data_iter = lambda: (s for s in calib_data)
    calib_data_tf = tf.data.Dataset.from_generator(
        data_iter,
        output_signature=(
            (tf.TensorSpec(shape=(None, None, len(calib_data.expanded_feature_names)), dtype=tf.float32),
             tf.TensorSpec(shape=(None, None), dtype=tf.bool)),
            tf.TensorSpec(shape=(None), dtype=tf.bool)
        ))

    calib_data_tf = calib_data_tf.prefetch(4 * num_gpus)
    options = tf.data.Options()
    options.experimental_distribute.auto_shard_policy = tf.data.experimental.AutoShardPolicy.DATA
    calib_data_tf = calib_data_tf.with_options(options)  # avoids Tensorflow ugly console barf
    calib_data_flat_y = model.predict(x=calib_data_tf,
                                     workers=args.n_procs,
                                     use_multiprocessing=True,
                                     max_queue_size=max(args.n_procs, 10),
                                     verbose=0)
    calib_data_predicted_score = calib_data.group(calib_data_flat_y, max) #it is anyway only one score
    auc_calib = average_precision_score(calib_data_y, calib_data_predicted_score)
    recall1_calib = recall_score(calib_data_y, calib_data_predicted_score > 0.5, pos_label=1)
    recall0_calib = recall_score(calib_data_y, calib_data_predicted_score > 0.5, pos_label=0)
    logging.info(f'Prediction scores for calib: aucPR: {auc_calib} - recall1: {recall1_calib} - recall0: {recall0_calib}')
    
    #fit isotonic regression
    iso_reg = IsotonicRegression(y_min=0, y_max=1, out_of_bounds='clip').fit(calib_data_predicted_score, calib_data_y)
    
    #for valid part predict with normal model
    predict_data = Models.BinaryDatasetEval(reader, eval_idx, args.features, args.max_len, max(250, args.max_len), 
                                            int(args.gpu_eval_mem_gb * 1e9 * 0.8), cache_results=False,
                                            show_progress=True, convoluted_size=convoluted_size,
                                            pad_to_max_len=is_fixed_length, batch_size=args.batch_size)

    eval_data_y = np.array([0 if reader.contigs[idx].misassembly == 0 else 1 for idx in predict_data.indices])
    # convert the slow Keras predict_data of type Sequence to a tf.data object
    data_iter = lambda: (s for s in predict_data)
    predict_data_tf = tf.data.Dataset.from_generator(
        data_iter,
        output_signature=(
            # first dimension is batch size, second is contig length, third is number of features
            (tf.TensorSpec(shape=(None, None, len(predict_data.expanded_feature_names)), dtype=tf.float32),
             # first dimension is batch size, second is contig length (no third dimension,
             # as all features are masked the same way)
             tf.TensorSpec(shape=(None, None), dtype=tf.bool)),
            tf.TensorSpec(shape=(None), dtype=tf.bool)
        ))
    predict_data_tf = predict_data_tf.prefetch(4 * num_gpus)
    options = tf.data.Options()
    options.experimental_distribute.auto_shard_policy = tf.data.experimental.AutoShardPolicy.DATA
    predict_data_tf = predict_data_tf.with_options(options)  # avoids Tensorflow ugly console barf
    eval_data_flat_y = model.predict(x=predict_data_tf,
                                     workers=args.n_procs,
                                     use_multiprocessing=True,
                                     max_queue_size=max(args.n_procs, 10),
                                     verbose=0)
    #aggregate with max
    eval_data_predicted_max = predict_data.group(eval_data_flat_y, max)
    auc_val = average_precision_score(eval_data_y, eval_data_predicted_max)
    recall1_val = recall_score(eval_data_y, eval_data_predicted_max > 0.5, pos_label=1)
    recall0_val = recall_score(eval_data_y, eval_data_predicted_max > 0.5, pos_label=0)
    logging.info(f'Eval data with max: aucPR: {auc_val} - recall1: {recall1_val} - recall0: {recall0_val}')
    
    #aggregate with isotonic + prob
    eval_data_calib_prob = predict_data.group(eval_data_flat_y, 'calib_prob', model=iso_reg)
    auc_calib_prob = average_precision_score(eval_data_y, eval_data_calib_prob)
    recall1_calib_prob = recall_score(eval_data_y, eval_data_calib_prob > 0.5, pos_label=1)
    recall0_calib_prob = recall_score(eval_data_y, eval_data_calib_prob > 0.5, pos_label=0)
    logging.info(f'Eval data with calib_prob: aucPR: {auc_calib_prob} - recall1: {recall1_calib_prob} - recall0: {recall0_calib_prob}')
    
    duration = time.time() - start
    logging.info(f'Prediction done in {duration:.0f}s.')

    # saving output
    if not os.path.isdir(args.save_path):
        os.makedirs(args.save_path)
    out_file = os.path.join(args.save_path, args.save_name + '.csv')    
   

    with open(out_file, 'w') as outF:
        outF.write('cont_name,length,label,calib_prob,max\n')
        for idx in range(len(eval_idx)):
            contig = reader.contigs[predict_data.indices[idx]]
            outF.write(f'{os.path.join(os.path.dirname(contig.file), contig.name)},{contig.length},'
                       f'{contig.misassembly},{eval_data_calib_prob[idx]},'
                       f'{eval_data_predicted_max[idx]}\n')

    logging.info(f'Predictions saved to: {out_file}')


#!/usr/bin/env python
from setuptools import setup, find_packages, Extension
from Cython.Build import cythonize
import platform
import numpy

# dependencies
install_reqs = [
    'cmake>=3.13',
    'ipython',
    'keras==2.8.0',
    'numpy>=1.17.0',
    'pandas==1.4.2',
    'pathos==0.2.9',
    'pytest==7.1.2',
    'snakemake==5.31.1',
    'tables',
    'pysam==0.19.1',
    'protobuf==3.20',
    'tensorflow==2.8.1',
    'tensorboard<2.9.0',
    'scikit-learn==1.1.1',
    'toolz==0.11.2'
]

compile_args = ['-std=c++11']
link_args = ['-std=c++11']
if platform.system() != 'Darwin':
    compile_args.append('-fopenmp')
    link_args.append('-fopenmp')

sourcefiles = ['resmico/reader.pyx', 'resmico/contig_reader.cpp']

with open("README.md", 'r') as f:
    long_description = f.read()

extensions = [Extension('resmico.reader',
                        sourcefiles,
                        language="c++",
                        extra_compile_args=compile_args,
                        extra_link_args=link_args,
                        libraries=['z'],
                        )]


class get_numpy_include(object):
    """
    Returns Numpy's include path with lazy import.
    """

    def __str__(self):
        import numpy
        return numpy.get_include()


## install main application
desc = 'Increasing the quality of metagenome-assembled genomes with deep learning'
setup(
    name='resmico',
    version='1.2.2',
    description=desc,
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Nick Youngblut, Olga Mineeva, Daniel Danciu',
    author_email='nyoungb2@gmail.com',
    entry_points={
        'console_scripts': [
            'resmico = resmico.__main__:main'
        ]
    },
    ext_modules=cythonize(extensions, include_path=['resmico/', numpy.get_include()], annotate=True,
                          compiler_directives={'language_level': "3"}),
    install_requires=install_reqs,
    include_dirs=[get_numpy_include()],
    license="MIT license",
    packages=find_packages(),
    package_data={'resmico': ['reader.pyx', 'contig_reader.hpp', 'contig_reader.cpp',
                              'bam2feat', 'model/resmico.h5', 'model/stats_cov.json']},
    package_dir={'resmico': 'resmico'},
    python_requires='>=3.8',
    setup_requires=[
        # Setuptools 18.0 properly handles Cython extensions.
        'setuptools>=18.0',
        'cython',
        'numpy',
    ],
    url='https://github.com/leylabmpi/resmico'
)
